from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

import numpy as np
import optuna
import pandas as pd
from sklearn.preprocessing import StandardScaler

from ins_pricing.modelling.bayesopt.trainers.cv_utils import CVStrategyResolver
from ins_pricing.utils import GPUMemoryManager, get_logger, log_print

_logger = get_logger("ins_pricing.trainer")


def _log(*args, **kwargs) -> None:
    log_print(_logger, *args, **kwargs)


class TrainerCVPredictionMixin:
    def _clean_gpu(
        self,
        *,
        synchronize: bool = True,
        empty_cache: bool = True,
    ) -> None:
        """Clean up GPU memory using shared GPUMemoryManager."""
        GPUMemoryManager.clean(synchronize=synchronize, empty_cache=empty_cache)

    def _standardize_fold(self,
                          X_train: pd.DataFrame,
                          X_val: pd.DataFrame,
                          columns: Optional[List[str]] = None
                          ) -> Tuple[pd.DataFrame, pd.DataFrame, StandardScaler]:
        """Fit StandardScaler on the training fold and transform train/val features."""
        scaler = StandardScaler()
        cols = list(columns) if columns else list(X_train.columns)
        X_train_scaled = X_train.copy(deep=True)
        X_val_scaled = X_val.copy(deep=True)
        if cols:
            scaler.fit(X_train_scaled[cols])
            X_train_scaled[cols] = scaler.transform(X_train_scaled[cols])
            X_val_scaled[cols] = scaler.transform(X_val_scaled[cols])
        return X_train_scaled, X_val_scaled, scaler

    def _resolve_train_val_indices(
        self,
        X_all: pd.DataFrame,
        *,
        allow_default: bool = False,
    ) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        """Resolve train/validation split indices based on configured CV strategy."""
        val_ratio = float(self.ctx.prop_test) if self.ctx.prop_test is not None else 0.25
        if not (0.0 < val_ratio < 1.0):
            if not allow_default:
                return None
            val_ratio = 0.25
        if len(X_all) < 10:
            return None

        resolver = CVStrategyResolver(self.ctx.config, self.ctx.train_data, self.ctx.rand_seed)
        (train_idx, val_idx), _ = resolver.create_train_val_splitter(X_all, val_ratio)
        return train_idx, val_idx

    def _resolve_time_sample_indices(
        self,
        X_all: pd.DataFrame,
        sample_limit: int,
    ) -> Optional[pd.Index]:
        """Get the most recent indices for time-based sampling."""
        if sample_limit <= 0:
            return None

        resolver = CVStrategyResolver(self.ctx.config, self.ctx.train_data, self.ctx.rand_seed)
        if not resolver.is_time_strategy():
            return None

        order = resolver.get_time_ordered_indices(X_all)
        if len(order) == 0:
            return None

        if len(order) > sample_limit:
            order = order[-sample_limit:]

        return X_all.index[order]

    def _resolve_ensemble_splits(
        self,
        X_all: pd.DataFrame,
        *,
        k: int,
    ) -> Tuple[Optional[Iterable[Tuple[np.ndarray, np.ndarray]]], int]:
        """Resolve K-fold splits for ensemble training based on configured CV strategy."""
        resolver = CVStrategyResolver(self.ctx.config, self.ctx.train_data, self.ctx.rand_seed)
        return resolver.create_kfold_splitter(X_all, k)

    def cross_val_generic(
            self,
            trial: optuna.trial.Trial,
            hyperparameter_space: Dict[str, Callable[[optuna.trial.Trial], Any]],
            data_provider: Callable[[], Tuple[pd.DataFrame, pd.Series, Optional[pd.Series]]],
            model_builder: Callable[[Dict[str, Any]], Any],
            metric_fn: Callable[[pd.Series, np.ndarray, Optional[pd.Series]], float],
            sample_limit: Optional[int] = None,
            preprocess_fn: Optional[Callable[[
                pd.DataFrame, pd.DataFrame], Tuple[pd.DataFrame, pd.DataFrame]]] = None,
            fit_predict_fn: Optional[
                Callable[[Any, pd.DataFrame, pd.Series, Optional[pd.Series],
                          pd.DataFrame, pd.Series, Optional[pd.Series],
                          optuna.trial.Trial], np.ndarray]
            ] = None,
            cleanup_fn: Optional[Callable[[Any], None]] = None,
            splitter: Optional[Iterable[Tuple[np.ndarray, np.ndarray]]] = None) -> float:
        """Generic holdout/CV helper to reuse tuning workflows."""
        params: Optional[Dict[str, Any]] = None
        if self._distributed_forced_params is not None:
            params = self._distributed_forced_params
            self._distributed_forced_params = None
        else:
            if trial is None:
                raise RuntimeError(
                    "Missing Optuna trial for parameter sampling.")
            params = {name: sampler(trial)
                      for name, sampler in hyperparameter_space.items()}
            if self._should_use_distributed_optuna():
                self._distributed_prepare_trial(params)
        X_all, y_all, w_all = data_provider()
        ctx_config = getattr(self.ctx, "config", None)
        cfg_limit = getattr(ctx_config, "bo_sample_limit", None)
        if cfg_limit is not None:
            cfg_limit = int(cfg_limit)
            if cfg_limit > 0:
                sample_limit = cfg_limit if sample_limit is None else min(sample_limit, cfg_limit)
        if sample_limit is not None and len(X_all) > sample_limit:
            sampled_idx = self._resolve_time_sample_indices(X_all, int(sample_limit))
            if sampled_idx is None:
                sampled_idx = X_all.sample(
                    n=sample_limit,
                    random_state=self.ctx.rand_seed
                ).index
            X_all = X_all.loc[sampled_idx]
            y_all = y_all.loc[sampled_idx]
            w_all = w_all.loc[sampled_idx] if w_all is not None else None

        if splitter is None:
            val_ratio = float(self.ctx.prop_test) if self.ctx.prop_test is not None else 0.25
            if not (0.0 < val_ratio < 1.0):
                val_ratio = 0.25
            cv_splits = getattr(self.ctx.config, "cv_splits", None)
            if cv_splits is None:
                cv_splits = max(2, int(round(1 / val_ratio)))
            cv_splits = max(2, int(cv_splits))

            resolver = CVStrategyResolver(self.ctx.config, self.ctx.train_data, self.ctx.rand_seed)
            split_iter, actual_splits = resolver.create_cv_splitter(X_all, y_all, cv_splits, val_ratio)
            if actual_splits < 2:
                raise ValueError("Not enough samples for cross-validation.")
        else:
            if hasattr(splitter, "split"):
                split_iter = splitter.split(X_all, y_all, groups=None)
            else:
                split_iter = splitter

        losses: List[float] = []
        for fold_idx, (train_idx, val_idx) in enumerate(split_iter):
            X_train = X_all.iloc[train_idx]
            y_train = y_all.iloc[train_idx]
            X_val = X_all.iloc[val_idx]
            y_val = y_all.iloc[val_idx]
            w_train = w_all.iloc[train_idx] if w_all is not None else None
            w_val = w_all.iloc[val_idx] if w_all is not None else None

            if preprocess_fn:
                X_train, X_val = preprocess_fn(X_train, X_val)

            model = model_builder(params)
            try:
                if fit_predict_fn:
                    trial_for_fold = trial if fold_idx == 0 else None
                    y_pred = fit_predict_fn(
                        model, X_train, y_train, w_train,
                        X_val, y_val, w_val, trial_for_fold
                    )
                else:
                    fit_kwargs = {}
                    if w_train is not None:
                        fit_kwargs["sample_weight"] = w_train
                    model.fit(X_train, y_train, **fit_kwargs)
                    y_pred = model.predict(X_val)
                losses.append(metric_fn(y_val, y_pred, w_val))
            finally:
                if cleanup_fn:
                    cleanup_fn(model)
                self._clean_gpu()

        return float(np.mean(losses))

    def _predict_and_cache(self,
                           model,
                           pred_prefix: str,
                           use_oht: bool = False,
                           design_fn=None,
                           predict_kwargs_train: Optional[Dict[str, Any]] = None,
                           predict_kwargs_test: Optional[Dict[str, Any]] = None,
                           predict_fn: Optional[Callable[..., Any]] = None) -> None:
        if design_fn:
            X_train = design_fn(train=True)
            X_test = design_fn(train=False)
        elif use_oht:
            X_train = self.ctx.train_oht_scl_data[self.ctx.var_nmes]
            X_test = self.ctx.test_oht_scl_data[self.ctx.var_nmes]
        else:
            X_train = self.ctx.train_data[self.ctx.factor_nmes]
            X_test = self.ctx.test_data[self.ctx.factor_nmes]

        predictor = predict_fn or model.predict
        preds_train = predictor(X_train, **(predict_kwargs_train or {}))
        preds_test = predictor(X_test, **(predict_kwargs_test or {}))
        preds_train, preds_test = self._store_predictions(
            pred_prefix, preds_train, preds_test
        )
        self._maybe_cache_predictions(pred_prefix, preds_train, preds_test)

    def _normalize_prediction_arrays(
        self,
        pred_prefix: str,
        preds_train: Any,
        preds_test: Any,
    ) -> Tuple[np.ndarray, np.ndarray, bool]:
        train_arr = np.asarray(preds_train)
        test_arr = np.asarray(preds_test)
        is_scalar = train_arr.ndim <= 1 or (
            train_arr.ndim == 2 and train_arr.shape[1] == 1
        )
        if is_scalar:
            return train_arr.reshape(-1), test_arr.reshape(-1), False
        if train_arr.ndim != 2:
            raise ValueError(
                f"Unexpected prediction shape for '{pred_prefix}': {train_arr.shape}"
            )
        if test_arr.ndim != 2 or test_arr.shape[1] != train_arr.shape[1]:
            raise ValueError(
                f"Train/test prediction dims mismatch for '{pred_prefix}': "
                f"{train_arr.shape} vs {test_arr.shape}"
            )
        return train_arr, test_arr, True

    def _store_predictions(self,
                           pred_prefix: str,
                           preds_train: Any,
                           preds_test: Any) -> Tuple[np.ndarray, np.ndarray]:
        train_arr, test_arr, is_vector = self._normalize_prediction_arrays(
            pred_prefix, preds_train, preds_test
        )
        if is_vector:
            self._assign_vector_predictions(pred_prefix, train_arr, test_arr)
            return train_arr, test_arr

        col_name = f'pred_{pred_prefix}'
        self.ctx.train_data[col_name] = train_arr
        self.ctx.test_data[col_name] = test_arr
        self.ctx.train_data[f'w_{col_name}'] = (
            self.ctx.train_data[col_name] *
            self.ctx.train_data[self.ctx.weight_nme]
        )
        self.ctx.test_data[f'w_{col_name}'] = (
            self.ctx.test_data[col_name] *
            self.ctx.test_data[self.ctx.weight_nme]
        )
        return train_arr, test_arr

    def _assign_vector_predictions(self,
                                   pred_prefix: str,
                                   preds_train: np.ndarray,
                                   preds_test: np.ndarray) -> None:
        col_names = [f'pred_{pred_prefix}_{j}' for j in range(preds_train.shape[1])]
        train_block = pd.DataFrame(
            preds_train, columns=col_names, index=self.ctx.train_data.index)
        test_block = pd.DataFrame(
            preds_test, columns=col_names, index=self.ctx.test_data.index)

        self.ctx.train_data = pd.concat(
            [self.ctx.train_data.drop(columns=col_names, errors='ignore'), train_block],
            axis=1
        )
        self.ctx.test_data = pd.concat(
            [self.ctx.test_data.drop(columns=col_names, errors='ignore'), test_block],
            axis=1
        )

    def _cache_predictions(self,
                           pred_prefix: str,
                           preds_train,
                           preds_test) -> None:
        preds_train, preds_test = self._store_predictions(
            pred_prefix, preds_train, preds_test
        )
        self._maybe_cache_predictions(pred_prefix, preds_train, preds_test)

    def _maybe_cache_predictions(self, pred_prefix: str, preds_train, preds_test) -> None:
        cfg = getattr(self.ctx, "config", None)
        if cfg is None or not bool(getattr(cfg, "cache_predictions", False)):
            return
        fmt = str(getattr(cfg, "prediction_cache_format", "parquet") or "parquet").lower()
        cache_dir = getattr(cfg, "prediction_cache_dir", None)
        if cache_dir:
            target_dir = Path(str(cache_dir))
            if not target_dir.is_absolute():
                target_dir = Path(self.output.result_dir) / target_dir
        else:
            target_dir = Path(self.output.result_dir) / "predictions"
        target_dir.mkdir(parents=True, exist_ok=True)

        def _build_frame(preds) -> pd.DataFrame:
            arr = np.asarray(preds)
            if arr.ndim <= 1:
                return pd.DataFrame({f"pred_{pred_prefix}": arr.reshape(-1)})
            cols = [f"pred_{pred_prefix}_{i}" for i in range(arr.shape[1])]
            return pd.DataFrame(arr, columns=cols)

        for split_label, preds in [("train", preds_train), ("test", preds_test)]:
            frame = _build_frame(preds)
            filename = f"{self.ctx.model_nme}_{pred_prefix}_{split_label}.{ 'csv' if fmt == 'csv' else 'parquet' }"
            path = target_dir / filename
            try:
                if fmt == "csv":
                    frame.to_csv(path, index=False)
                else:
                    frame.to_parquet(path, index=False)
            except Exception:
                pass

    def _resolve_best_epoch(self,
                            history: Optional[Dict[str, List[float]]],
                            default_epochs: int) -> int:
        if not history:
            return max(1, int(default_epochs))
        vals = history.get("val") or []
        if not vals:
            return max(1, int(default_epochs))
        best_idx = int(np.nanargmin(vals))
        return max(1, best_idx + 1)

    def _fit_predict_cache(self,
                           model,
                           X_train,
                           y_train,
                           sample_weight,
                           pred_prefix: str,
                           use_oht: bool = False,
                           design_fn=None,
                           fit_kwargs: Optional[Dict[str, Any]] = None,
                           sample_weight_arg: Optional[str] = 'sample_weight',
                           predict_kwargs_train: Optional[Dict[str, Any]] = None,
                           predict_kwargs_test: Optional[Dict[str, Any]] = None,
                           predict_fn: Optional[Callable[..., Any]] = None,
                           record_label: bool = True) -> None:
        fit_kwargs = fit_kwargs.copy() if fit_kwargs else {}
        if sample_weight is not None and sample_weight_arg:
            fit_kwargs.setdefault(sample_weight_arg, sample_weight)
        model.fit(X_train, y_train, **fit_kwargs)
        if record_label:
            self.ctx.model_label.append(self.label)
        self._predict_and_cache(
            model,
            pred_prefix,
            use_oht=use_oht,
            design_fn=design_fn,
            predict_kwargs_train=predict_kwargs_train,
            predict_kwargs_test=predict_kwargs_test,
            predict_fn=predict_fn)


__all__ = ["TrainerCVPredictionMixin"]



"""
Unified Task Runner with Real-time Logging
Executes model training, explanation, plotting, and other tasks based on config.
"""

import sys
import os
import threading
import queue
import time
import json
import subprocess
import signal
from pathlib import Path
from typing import Generator, Optional, Dict, Any, List, Sequence, Tuple
import logging

from ins_pricing.utils import get_logger, log_print

_logger = get_logger("ins_pricing.frontend.runner")


def _log(*args, **kwargs) -> None:
    log_print(_logger, *args, **kwargs)

class LogCapture:
    """Capture stdout and stderr for real-time display."""

    def __init__(self):
        self.log_queue = queue.Queue()
        self.stop_flag = threading.Event()

    def write(self, text: str):
        """Write method for capturing output."""
        if text and text.strip():
            self.log_queue.put(text)

    def flush(self):
        """Flush method (required for file-like objects)."""
        pass


class TaskRunner:
    """
    Run model tasks (training, explain, plotting, etc.) and capture logs.

    Supports all task modes defined in config.runner.mode:
    - entry: Standard model training
    - explain: Model explanation (permutation, SHAP, etc.)
    - incremental: Incremental training
    - watchdog: Watchdog mode for monitoring
    """

    def __init__(self):
        self.task_thread = None
        self.log_capture = None
        self._proc: Optional[subprocess.Popen] = None

    def _detect_task_mode(self, config_path: str) -> str:
        """
        Detect the task mode from config file.

        Args:
            config_path: Path to configuration JSON file

        Returns:
            Task mode string (e.g., 'entry', 'explain', 'incremental', 'watchdog')
        """
        try:
            with open(config_path, 'r', encoding='utf-8') as f:
                config = json.load(f)

            runner_config = config.get('runner', {})
            mode = runner_config.get('mode', 'entry')
            return str(mode).lower()

        except Exception as e:
            _log(f"Warning: Could not detect task mode, defaulting to 'entry': {e}")
            return 'entry'

    def _build_cmd_from_config(self, config_path: str) -> Tuple[List[str], str]:
        """Build the command to execute based on config.runner.mode."""
        from ins_pricing.cli.utils.notebook_utils import build_cmd_from_config

        return build_cmd_from_config(config_path)

    def run_task(self, config_path: str) -> Generator[str, None, None]:
        """
        Run task based on config file with real-time log capture.

        This method automatically detects the task mode from the config file
        (training, explain, plotting, etc.) and runs the appropriate task.

        Args:
            config_path: Path to configuration JSON file

        Yields:
            Log lines as they are generated
        """
        self.log_capture = LogCapture()

        # Configure logging to capture both file and stream output
        log_handler = logging.StreamHandler(self.log_capture)
        log_handler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        log_handler.setFormatter(formatter)

        # Add handler to root logger
        root_logger = logging.getLogger()
        original_handlers = root_logger.handlers.copy()
        root_logger.addHandler(log_handler)

        # Store original stdout/stderr
        original_stdout = sys.stdout
        original_stderr = sys.stderr

        try:
            # Detect task mode
            task_mode = self._detect_task_mode(config_path)

            # Start task in separate thread
            exception_holder = []

            def task_worker():
                try:
                    sys.stdout = self.log_capture
                    sys.stderr = self.log_capture

                    # Log start
                    cmd, task_mode = self._build_cmd_from_config(config_path)
                    _log(f"Starting task [{task_mode}] with config: {config_path}")
                    _log("=" * 80)

                    # Run subprocess with streamed output
                    creationflags = 0
                    start_new_session = False
                    if os.name == "nt":
                        creationflags = getattr(subprocess, "CREATE_NEW_PROCESS_GROUP", 0)
                    else:
                        start_new_session = True
                    proc = subprocess.Popen(
                        cmd,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        text=True,
                        bufsize=1,
                        cwd=str(Path(config_path).resolve().parent),
                        creationflags=creationflags,
                        start_new_session=start_new_session,
                    )
                    self._proc = proc
                    if proc.stdout is not None:
                        for line in proc.stdout:
                            _log(line.rstrip())
                    return_code = proc.wait()
                    if return_code != 0:
                        raise RuntimeError(f"Task exited with code {return_code}")

                    _log("=" * 80)
                    _log(f"Task [{task_mode}] completed successfully!")

                except Exception as e:
                    exception_holder.append(e)
                    _log(f"Error during task execution: {str(e)}")
                    import traceback
                    _log(traceback.format_exc())

                finally:
                    self._proc = None
                    sys.stdout = original_stdout
                    sys.stderr = original_stderr

            self.task_thread = threading.Thread(target=task_worker, daemon=True)
            self.task_thread.start()

            # Yield log lines as they come in
            last_update = time.time()
            while self.task_thread.is_alive() or not self.log_capture.log_queue.empty():
                try:
                    # Try to get log with timeout
                    log_line = self.log_capture.log_queue.get(timeout=0.1)
                    yield log_line
                    last_update = time.time()

                except queue.Empty:
                    # Send heartbeat every 5 seconds
                    if time.time() - last_update > 5:
                        yield "."
                        last_update = time.time()
                    continue

            # Wait for thread to complete
            self.task_thread.join(timeout=1)

            # Check for exceptions
            if exception_holder:
                raise exception_holder[0]

        finally:
            # Restore original stdout/stderr
            sys.stdout = original_stdout
            sys.stderr = original_stderr

            # Restore original logging handlers
            root_logger.handlers = original_handlers

    def run_callable(self, func, *args, **kwargs) -> Generator[str, None, None]:
        """Run an in-process callable and stream stdout/stderr."""
        self.log_capture = LogCapture()

        log_handler = logging.StreamHandler(self.log_capture)
        log_handler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        log_handler.setFormatter(formatter)

        root_logger = logging.getLogger()
        original_handlers = root_logger.handlers.copy()
        root_logger.addHandler(log_handler)

        original_stdout = sys.stdout
        original_stderr = sys.stderr

        try:
            exception_holder = []

            def task_worker():
                try:
                    sys.stdout = self.log_capture
                    sys.stderr = self.log_capture
                    func(*args, **kwargs)
                except Exception as e:
                    exception_holder.append(e)
                    _log(f"Error during task execution: {str(e)}")
                    import traceback
                    _log(traceback.format_exc())
                finally:
                    sys.stdout = original_stdout
                    sys.stderr = original_stderr

            self.task_thread = threading.Thread(target=task_worker, daemon=True)
            self.task_thread.start()

            last_update = time.time()
            while self.task_thread.is_alive() or not self.log_capture.log_queue.empty():
                try:
                    log_line = self.log_capture.log_queue.get(timeout=0.1)
                    yield log_line
                    last_update = time.time()
                except queue.Empty:
                    if time.time() - last_update > 5:
                        yield "."
                        last_update = time.time()
                    continue

            self.task_thread.join(timeout=1)
            if exception_holder:
                raise exception_holder[0]
        finally:
            sys.stdout = original_stdout
            sys.stderr = original_stderr
            root_logger.handlers = original_handlers

    def stop_task(self):
        """Stop the current task process."""
        if self.log_capture:
            self.log_capture.stop_flag.set()

        proc = self._proc
        if proc is not None and proc.poll() is None:
            try:
                if os.name == "nt":
                    subprocess.run(
                        ["taskkill", "/PID", str(proc.pid), "/T", "/F"],
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                        check=False,
                    )
                else:
                    try:
                        os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
                    except Exception:
                        proc.terminate()
                try:
                    proc.wait(timeout=5)
                except Exception:
                    if os.name == "nt":
                        proc.kill()
                    else:
                        try:
                            os.killpg(os.getpgid(proc.pid), signal.SIGKILL)
                        except Exception:
                            proc.kill()
            except Exception:
                try:
                    if os.name != "nt":
                        try:
                            os.killpg(os.getpgid(proc.pid), signal.SIGKILL)
                        except Exception:
                            proc.kill()
                    else:
                        proc.kill()
                except Exception:
                    pass

        if self.task_thread and self.task_thread.is_alive():
            self.task_thread.join(timeout=5)


# Backward compatibility aliases
TrainingRunner = TaskRunner


class StreamToLogger:
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """

    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ''

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())

    def flush(self):
        pass


def setup_logger(name: str = "task") -> logging.Logger:
    """
    Set up a logger for task execution.

    Args:
        name: Logger name

    Returns:
        Configured logger instance
    """
    logger_name = name if "." in name else f"ins_pricing.frontend.{name}"
    logger = get_logger(logger_name)
    logger.setLevel(logging.INFO)
    return logger



"""Tests for pricing exposure calculation module."""

import numpy as np
import pandas as pd
import pytest
from datetime import datetime, timedelta


@pytest.fixture
def sample_policy_dates():
    """Sample policy with start and end dates."""
    start_date = datetime(2023, 1, 1)
    return pd.DataFrame({
        "policy_id": range(100),
        "start_date": [start_date + timedelta(days=i) for i in range(100)],
        "end_date": [start_date + timedelta(days=i+365) for i in range(100)],
        "premium": np.random.uniform(200, 1000, 100)
    })


class TestExposureCalculation:
    """Test exposure calculation functions."""

    def test_calculate_policy_exposure_years(self, sample_policy_dates):
        """Test calculating exposure in years."""
        from ins_pricing.pricing.exposure import compute_exposure

        df = compute_exposure(
            sample_policy_dates,
            start_col="start_date",
            end_col="end_date",
            time_unit="years"
        )

        assert "exposure" in df.columns
        assert all(df["exposure"] > 0)
        assert all(df["exposure"] <= 1.1)  # Roughly 1 year

    def test_calculate_policy_exposure_days(self, sample_policy_dates):
        """Test calculating exposure in days."""
        from ins_pricing.pricing.exposure import compute_exposure

        df = compute_exposure(
            sample_policy_dates,
            start_col="start_date",
            end_col="end_date",
            time_unit="days"
        )

        assert all(df["exposure"] >= 365)
        assert all(df["exposure"] <= 366)

    def test_partial_period_exposure(self):
        """Test exposure for partial periods."""
        from ins_pricing.pricing.exposure import compute_exposure

        df = pd.DataFrame({
            "start_date": [datetime(2023, 1, 1)],
            "end_date": [datetime(2023, 6, 30)]  # 6 months
        })

        result = compute_exposure(df, "start_date", "end_date", time_unit="years")

        assert 0.48 < result["exposure"].iloc[0] < 0.52  # Roughly 0.5 years



"""BayesOpt subpackage (split from monolithic BayesOpt.py)."""

from __future__ import annotations

from ins_pricing.utils.torch_compat import disable_torch_dynamo_if_requested

disable_torch_dynamo_if_requested()

from ins_pricing.modelling.bayesopt.config_preprocess import (
    BayesOptConfig,
    DatasetPreprocessor,
    OutputManager,
    VersionManager,
)
from ins_pricing.modelling.bayesopt.core import BayesOptModel
from ins_pricing.modelling.bayesopt.models import (
    FeatureTokenizer,
    FTTransformerCore,
    FTTransformerSklearn,
    GraphNeuralNetSklearn,
    MaskedTabularDataset,
    ResBlock,
    ResNetSequential,
    ResNetSklearn,
    ScaledTransformerEncoderLayer,
    SimpleGraphLayer,
    SimpleGNN,
    TabularDataset,
)
from ins_pricing.modelling.bayesopt.trainers import (
    FTTrainer,
    GLMTrainer,
    GNNTrainer,
    ResNetTrainer,
    TrainerBase,
    XGBTrainer,
    _xgb_cuda_available,
)
__all__ = [
    "BayesOptConfig",
    "DatasetPreprocessor",
    "OutputManager",
    "VersionManager",
    "BayesOptModel",
    "FeatureTokenizer",
    "FTTransformerCore",
    "FTTransformerSklearn",
    "GraphNeuralNetSklearn",
    "MaskedTabularDataset",
    "ResBlock",
    "ResNetSequential",
    "ResNetSklearn",
    "ScaledTransformerEncoderLayer",
    "SimpleGraphLayer",
    "SimpleGNN",
    "TabularDataset",
    "FTTrainer",
    "GLMTrainer",
    "GNNTrainer",
    "ResNetTrainer",
    "TrainerBase",
    "XGBTrainer",
    "_xgb_cuda_available",
]



"""Unified import resolver for CLI modules.

This module provides a single source of truth for all import fallback chains,
eliminating the need for nested try/except blocks in multiple CLI files.

Usage:
    from ins_pricing.cli.utils.import_resolver import resolve_imports
    imports = resolve_imports()
    ropt = imports.bayesopt
    PLOT_MODEL_LABELS = imports.PLOT_MODEL_LABELS
"""

from __future__ import annotations

import importlib
import os
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from ins_pricing.cli.utils.bootstrap import ensure_repo_root


@dataclass
class ResolvedImports:
    """Container for resolved imports from the bayesopt ecosystem."""

    # Core bayesopt module
    bayesopt: Any = None

    # CLI common utilities
    PLOT_MODEL_LABELS: Dict[str, Tuple[str, str]] = field(default_factory=dict)
    PYTORCH_TRAINERS: List[str] = field(default_factory=list)
    build_model_names: Optional[Callable] = None
    dedupe_preserve_order: Optional[Callable] = None
    load_dataset: Optional[Callable] = None
    parse_model_pairs: Optional[Callable] = None
    resolve_data_path: Optional[Callable] = None
    resolve_path: Optional[Callable] = None
    fingerprint_file: Optional[Callable] = None
    coerce_dataset_types: Optional[Callable] = None
    split_train_test: Optional[Callable] = None

    # CLI config utilities
    add_config_json_arg: Optional[Callable] = None
    add_output_dir_arg: Optional[Callable] = None
    resolve_and_load_config: Optional[Callable] = None
    resolve_data_config: Optional[Callable] = None
    resolve_report_config: Optional[Callable] = None
    resolve_explain_save_root: Optional[Callable] = None
    resolve_explain_save_dir: Optional[Callable] = None
    resolve_explain_output_overrides: Optional[Callable] = None
    resolve_model_path_value: Optional[Callable] = None
    resolve_split_config: Optional[Callable] = None
    resolve_runtime_config: Optional[Callable] = None
    resolve_output_dirs: Optional[Callable] = None

    # Evaluation utilities
    bootstrap_ci: Optional[Callable] = None
    calibrate_predictions: Optional[Callable] = None
    metrics_report: Optional[Callable] = None
    select_threshold: Optional[Callable] = None

    # Governance and reporting
    ModelArtifact: Optional[Type] = None
    ModelRegistry: Optional[Type] = None
    drift_psi_report: Optional[Callable] = None
    group_metrics: Optional[Callable] = None
    ReportPayload: Optional[Type] = None
    write_report: Optional[Callable] = None

    # Logging
    configure_run_logging: Optional[Callable] = None

    # Plotting
    plot_loss_curve: Optional[Callable] = None


def _debug_imports_enabled() -> bool:
    value = os.environ.get("BAYESOPT_DEBUG_IMPORTS")
    if value is None:
        return False
    return str(value).strip().lower() in {"1", "true", "yes", "y", "on"}


def _try_import(module_path: str, attr_name: Optional[str] = None) -> Optional[Any]:
    """Attempt to import a module or attribute, returning None on failure."""
    try:
        module = importlib.import_module(module_path)
        if attr_name:
            result = getattr(module, attr_name, None)
        else:
            result = module
        if _debug_imports_enabled():
            origin = getattr(module, "__file__", None)
            origin = origin or getattr(module, "__path__", None)
            print(
                f"[BAYESOPT_DEBUG_IMPORTS] imported {module_path}"
                f"{'::' + attr_name if attr_name else ''} from {origin}",
                file=sys.stderr,
                flush=True,
            )
        return result
    except Exception as exc:
        if _debug_imports_enabled():
            print(
                f"[BAYESOPT_DEBUG_IMPORTS] failed import {module_path}"
                f"{'::' + attr_name if attr_name else ''}: {exc.__class__.__name__}: {exc}",
                file=sys.stderr,
                flush=True,
            )
        return None


def _try_import_from_paths(
    paths: List[str],
    attr_name: Optional[str] = None
) -> Optional[Any]:
    """Try importing from multiple module paths, return first success."""
    for path in paths:
        result = _try_import(path, attr_name)
        if result is not None:
            return result
    return None


def _collect_attrs_from_paths(
    paths: List[str],
    attrs: List[str],
    *,
    min_found: int = 1,
) -> Dict[str, Any]:
    results: Dict[str, Any] = {attr: None for attr in attrs}
    for path in paths:
        module = _try_import(path)
        if module is None:
            continue
        for attr in attrs:
            if results[attr] is None:
                results[attr] = getattr(module, attr, None)
        found = sum(1 for value in results.values() if value is not None)
        if found >= min_found:
            break
    return results


def _resolve_bayesopt() -> Optional[Any]:
    """Resolve the bayesopt module."""
    paths = ["ins_pricing.modelling.bayesopt"]
    return _try_import_from_paths(paths)


def _resolve_cli_common() -> Dict[str, Any]:
    """Resolve CLI common utilities."""
    paths = ["ins_pricing.cli.utils.cli_common"]

    attrs = [
        "PLOT_MODEL_LABELS",
        "PYTORCH_TRAINERS",
        "build_model_names",
        "dedupe_preserve_order",
        "load_dataset",
        "parse_model_pairs",
        "resolve_data_path",
        "resolve_path",
        "fingerprint_file",
        "coerce_dataset_types",
        "split_train_test",
    ]

    return _collect_attrs_from_paths(
        paths,
        attrs,
        min_found=max(1, len(attrs) // 2),
    )


def _resolve_cli_config() -> Dict[str, Any]:
    """Resolve CLI config utilities."""
    paths = ["ins_pricing.cli.utils.cli_config"]

    attrs = [
        "add_config_json_arg",
        "add_output_dir_arg",
        "resolve_and_load_config",
        "resolve_data_config",
        "resolve_report_config",
        "resolve_explain_save_root",
        "resolve_explain_save_dir",
        "resolve_explain_output_overrides",
        "resolve_model_path_value",
        "resolve_split_config",
        "resolve_runtime_config",
        "resolve_output_dirs",
    ]

    return _collect_attrs_from_paths(
        paths,
        attrs,
        min_found=max(1, len(attrs) // 2),
    )


def _resolve_evaluation() -> Dict[str, Any]:
    """Resolve evaluation utilities."""
    paths = ["ins_pricing.modelling.evaluation"]

    return _collect_attrs_from_paths(
        paths,
        ["bootstrap_ci", "calibrate_predictions", "metrics_report", "select_threshold"],
        min_found=1,
    )


def _resolve_governance() -> Dict[str, Any]:
    """Resolve governance and reporting utilities."""
    results: Dict[str, Any] = {}
    results.update(
        _collect_attrs_from_paths(
            ["ins_pricing.governance.registry"],
            ["ModelArtifact", "ModelRegistry"],
            min_found=1,
        )
    )
    results.update(
        _collect_attrs_from_paths(
            ["ins_pricing.production"],
            ["drift_psi_report"],
            min_found=1,
        )
    )
    if results.get("drift_psi_report") is None:
        production_mod = _try_import("ins_pricing.production")
        if production_mod is not None:
            results["drift_psi_report"] = getattr(production_mod, "psi_report", None)
    results.update(
        _collect_attrs_from_paths(
            ["ins_pricing.production.monitoring"],
            ["group_metrics"],
            min_found=1,
        )
    )
    results.update(
        _collect_attrs_from_paths(
            ["ins_pricing.reporting.report_builder"],
            ["ReportPayload", "write_report"],
            min_found=1,
        )
    )
    return results


def _resolve_logging() -> Dict[str, Any]:
    """Resolve logging utilities."""
    paths = ["ins_pricing.cli.utils.run_logging"]

    return _collect_attrs_from_paths(paths, ["configure_run_logging"], min_found=1)


def _resolve_plotting() -> Dict[str, Any]:
    """Resolve plotting utilities."""
    paths = [
        "ins_pricing.modelling.plotting.diagnostics",
    ]

    return _collect_attrs_from_paths(paths, ["plot_loss_curve"], min_found=1)


def resolve_imports() -> ResolvedImports:
    """Resolve all imports from the bayesopt ecosystem.

    This function attempts to import modules from multiple possible locations,
    handling the various ways the package might be installed or run.

    Returns:
        ResolvedImports object containing all resolved imports.
    """
    imports = ResolvedImports()

    # Resolve bayesopt core
    imports.bayesopt = _resolve_bayesopt()

    # Resolve CLI common utilities
    cli_common = _resolve_cli_common()
    imports.PLOT_MODEL_LABELS = cli_common.get("PLOT_MODEL_LABELS", {})
    imports.PYTORCH_TRAINERS = cli_common.get("PYTORCH_TRAINERS", [])
    imports.build_model_names = cli_common.get("build_model_names")
    imports.dedupe_preserve_order = cli_common.get("dedupe_preserve_order")
    imports.load_dataset = cli_common.get("load_dataset")
    imports.parse_model_pairs = cli_common.get("parse_model_pairs")
    imports.resolve_data_path = cli_common.get("resolve_data_path")
    imports.resolve_path = cli_common.get("resolve_path")
    imports.fingerprint_file = cli_common.get("fingerprint_file")
    imports.coerce_dataset_types = cli_common.get("coerce_dataset_types")
    imports.split_train_test = cli_common.get("split_train_test")

    # Resolve CLI config utilities
    cli_config = _resolve_cli_config()
    imports.add_config_json_arg = cli_config.get("add_config_json_arg")
    imports.add_output_dir_arg = cli_config.get("add_output_dir_arg")
    imports.resolve_and_load_config = cli_config.get("resolve_and_load_config")
    imports.resolve_data_config = cli_config.get("resolve_data_config")
    imports.resolve_report_config = cli_config.get("resolve_report_config")
    imports.resolve_explain_save_root = cli_config.get("resolve_explain_save_root")
    imports.resolve_explain_save_dir = cli_config.get("resolve_explain_save_dir")
    imports.resolve_explain_output_overrides = cli_config.get("resolve_explain_output_overrides")
    imports.resolve_model_path_value = cli_config.get("resolve_model_path_value")
    imports.resolve_split_config = cli_config.get("resolve_split_config")
    imports.resolve_runtime_config = cli_config.get("resolve_runtime_config")
    imports.resolve_output_dirs = cli_config.get("resolve_output_dirs")

    # Resolve evaluation utilities
    evaluation = _resolve_evaluation()
    imports.bootstrap_ci = evaluation.get("bootstrap_ci")
    imports.calibrate_predictions = evaluation.get("calibrate_predictions")
    imports.metrics_report = evaluation.get("metrics_report")
    imports.select_threshold = evaluation.get("select_threshold")

    # Resolve governance and reporting
    governance = _resolve_governance()
    imports.ModelArtifact = governance.get("ModelArtifact")
    imports.ModelRegistry = governance.get("ModelRegistry")
    imports.drift_psi_report = governance.get("drift_psi_report")
    imports.group_metrics = governance.get("group_metrics")
    imports.ReportPayload = governance.get("ReportPayload")
    imports.write_report = governance.get("write_report")

    # Resolve logging
    logging_utils = _resolve_logging()
    imports.configure_run_logging = logging_utils.get("configure_run_logging")

    # Resolve plotting
    plotting = _resolve_plotting()
    imports.plot_loss_curve = plotting.get("plot_loss_curve")

    return imports


# Convenience function for backward compatibility
def setup_sys_path() -> None:
    """Ensure the repository root is in sys.path for imports."""
    try:
        if importlib.util.find_spec("ins_pricing") is not None:
            return
    except Exception:
        pass
    ensure_repo_root()



from __future__ import annotations

from importlib import import_module

_ROOT_SUBPACKAGES = {
    "modelling": "ins_pricing.modelling",
    "pricing": "ins_pricing.pricing",
    "production": "ins_pricing.production",
    "governance": "ins_pricing.governance",
    "reporting": "ins_pricing.reporting",
}

__all__ = sorted(set(_ROOT_SUBPACKAGES))


def __getattr__(name: str):
    target = _ROOT_SUBPACKAGES.get(name)
    if target is None:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    module = import_module(target)
    globals()[name] = module
    return module


def __dir__() -> list[str]:
    return sorted(set(__all__) | set(globals().keys()))



from __future__ import annotations

import types
from pathlib import Path

import pytest

optuna = pytest.importorskip("optuna")
pytest.importorskip("xgboost")

from ins_pricing.modelling.bayesopt.runtime.trainer_optuna import TrainerOptunaMixin
from ins_pricing.modelling.bayesopt.trainers.trainer_xgb import (
    XGBTrainer,
    _XGBDMatrixWrapper,
)


class _DummyTrainer(TrainerOptunaMixin):
    def __init__(self, tmp_path: Path) -> None:
        self.label = "Dummy"
        self.model_name_prefix = "dummy"
        self.best_params = None
        self.best_trial = None
        self.study_name = None
        self.enable_distributed_optuna = False
        self.ctx = types.SimpleNamespace(rand_seed=13, model_nme="demo_model")
        self.config = types.SimpleNamespace(
            optuna_storage=None,
            optuna_study_prefix="unit",
            optuna_cleanup_synchronize=False,
        )
        self.output = types.SimpleNamespace(result_dir=str(tmp_path))

    def _clean_gpu(self, synchronize: bool = False) -> None:
        return None


def test_run_study_and_extract_raises_for_all_pruned_trials(tmp_path: Path) -> None:
    trainer = _DummyTrainer(tmp_path)
    study = trainer._create_study()
    progress_counter = {"count": 0}

    def _always_prune(_trial: optuna.trial.Trial) -> float:
        raise optuna.TrialPruned("forced prune for test")

    objective_wrapper = trainer._make_objective_wrapper(
        _always_prune,
        total_trials=1,
        progress_counter=progress_counter,
        barrier_on_end=False,
    )
    checkpoint_callback = trainer._make_checkpoint_callback()

    with pytest.raises(RuntimeError, match="No completed trials"):
        trainer._run_study_and_extract(
            study=study,
            objective_wrapper=objective_wrapper,
            checkpoint_callback=checkpoint_callback,
            total_trials=1,
            progress_counter=progress_counter,
        )


def test_xgb_slow_prune_disabled_when_trial_budget_is_tiny() -> None:
    trainer = object.__new__(XGBTrainer)

    trainer._optuna_total_trials = 1
    assert trainer._should_prune_slow_config(max_depth=21, n_estimators=490) is False

    trainer._optuna_total_trials = 2
    assert trainer._should_prune_slow_config(max_depth=21, n_estimators=490) is False

    trainer._optuna_total_trials = 3
    assert trainer._should_prune_slow_config(max_depth=21, n_estimators=490) is True


def test_xgb_chunk_plan_covers_all_rows_and_rounds() -> None:
    plan = _XGBDMatrixWrapper._build_chunk_plan(
        total_rows=10,
        chunk_size=3,
        num_boost_round=5,
    )
    assert plan[0][0] == 0
    assert plan[-1][1] == 10
    assert sum(end - start for start, end, _ in plan) == 10
    assert sum(rounds for _, _, rounds in plan) == 5
    assert all(rounds >= 1 for _, _, rounds in plan)
    for prev, curr in zip(plan, plan[1:]):
        assert prev[1] == curr[0]


def test_xgb_chunk_plan_limits_chunk_count_by_boost_rounds() -> None:
    plan = _XGBDMatrixWrapper._build_chunk_plan(
        total_rows=100,
        chunk_size=10,
        num_boost_round=3,
    )
    assert len(plan) == 3
    assert sum(rounds for _, _, rounds in plan) == 3
    assert plan[-1][1] == 100


def test_run_study_and_extract_recovers_from_pruned_only_history(tmp_path: Path) -> None:
    trainer = _DummyTrainer(tmp_path)
    study = trainer._create_study()
    progress_counter = {"count": 0}

    def _always_prune(_trial: optuna.trial.Trial) -> float:
        raise optuna.TrialPruned("seed pruned trial")

    study.optimize(_always_prune, n_trials=1)

    def _complete(trial: optuna.trial.Trial) -> float:
        value = trial.suggest_float("learning_rate", 1e-5, 1e-2, log=True)
        return float(value)

    objective_wrapper = trainer._make_objective_wrapper(
        _complete,
        total_trials=1,
        progress_counter=progress_counter,
        barrier_on_end=False,
    )
    checkpoint_callback = trainer._make_checkpoint_callback()

    trainer._run_study_and_extract(
        study=study,
        objective_wrapper=objective_wrapper,
        checkpoint_callback=checkpoint_callback,
        total_trials=1,
        progress_counter=progress_counter,
    )

    assert isinstance(trainer.best_params, dict)
    assert "learning_rate" in trainer.best_params


def test_objective_wrapper_prunes_dataloader_worker_runtime_error(tmp_path: Path) -> None:
    trainer = _DummyTrainer(tmp_path)
    study = trainer._create_study()
    trial = study.ask()
    progress_counter = {"count": 0}

    def _raise_dataloader_runtime(_trial: optuna.trial.Trial) -> float:
        raise RuntimeError("DataLoader worker (pid 123) exited unexpectedly")

    objective_wrapper = trainer._make_objective_wrapper(
        _raise_dataloader_runtime,
        total_trials=1,
        progress_counter=progress_counter,
        barrier_on_end=False,
    )

    with pytest.raises(optuna.TrialPruned):
        objective_wrapper(trial)
    assert progress_counter["count"] == 1



"""Path resolution and data loading utilities.

This module consolidates path handling logic from:
- cli/utils/cli_common.py
- production/predict.py
- Various model loaders

Example:
    >>> from ins_pricing.utils import resolve_path, load_dataset
    >>> path = resolve_path("./data/train.csv", base_dir)
    >>> df = load_dataset(path, data_format="auto")
"""

from __future__ import annotations

import hashlib
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd


# =============================================================================
# Model Label Mappings
# =============================================================================

PLOT_MODEL_LABELS: Dict[str, Tuple[str, str]] = {
    "glm": ("GLM", "pred_glm"),
    "xgb": ("Xgboost", "pred_xgb"),
    "resn": ("ResNet", "pred_resn"),
    "ft": ("FTTransformer", "pred_ft"),
    "gnn": ("GNN", "pred_gnn"),
}

PYTORCH_TRAINERS = {"resn", "ft", "gnn"}


# =============================================================================
# List Utilities
# =============================================================================


def dedupe_preserve_order(items: Iterable[str]) -> List[str]:
    """Remove duplicates while preserving order.

    Args:
        items: Iterable of strings

    Returns:
        List with duplicates removed, original order preserved
    """
    seen = set()
    unique: List[str] = []
    for item in items:
        if item not in seen:
            unique.append(item)
            seen.add(item)
    return unique


def build_model_names(prefixes: Sequence[str], suffixes: Sequence[str]) -> List[str]:
    """Build model names from prefixes and suffixes.

    Args:
        prefixes: Model type prefixes (e.g., ['bi', 'od'])
        suffixes: Model category suffixes (e.g., ['bc', 'nc'])

    Returns:
        List of combined names (e.g., ['bi_bc', 'od_bc', 'bi_nc', 'od_nc'])
    """
    names: List[str] = []
    for suffix in suffixes:
        names.extend(f"{prefix}_{suffix}" for prefix in prefixes)
    return names


def parse_model_pairs(raw_pairs: List) -> List[Tuple[str, str]]:
    """Parse model pairs from various formats.

    Args:
        raw_pairs: List of pairs in various formats

    Returns:
        List of (model1, model2) tuples
    """
    pairs: List[Tuple[str, str]] = []
    for pair in raw_pairs:
        if isinstance(pair, (list, tuple)) and len(pair) == 2:
            pairs.append((str(pair[0]), str(pair[1])))
        elif isinstance(pair, str):
            parts = [p.strip() for p in pair.split(",") if p.strip()]
            if len(parts) == 2:
                pairs.append((parts[0], parts[1]))
    return pairs


# =============================================================================
# Path Resolution
# =============================================================================


def resolve_path(value: Optional[str], base_dir: Path) -> Optional[Path]:
    """Resolve a path relative to a base directory.

    Args:
        value: Path string (absolute or relative)
        base_dir: Base directory for relative paths

    Returns:
        Resolved absolute Path, or None if value is None/empty
    """
    if value is None:
        return None
    if not isinstance(value, str) or not value.strip():
        return None
    p = Path(value)
    if p.is_absolute():
        return p
    return (base_dir / p).resolve()


def resolve_dir_path(
    value: Optional[Union[str, Path]],
    base_dir: Path,
    *,
    create: bool = False,
) -> Optional[Path]:
    """Resolve a directory path, optionally creating it.

    Args:
        value: Path string or Path object
        base_dir: Base directory for relative paths
        create: Whether to create the directory if it doesn't exist

    Returns:
        Resolved absolute Path, or None if value is None/empty
    """
    if value is None:
        return None
    if isinstance(value, Path):
        path = value if value.is_absolute() else (base_dir / value).resolve()
    else:
        value_str = str(value)
        if not value_str.strip():
            return None
        path = resolve_path(value_str, base_dir)
        if path is None:
            path = Path(value_str)
            if not path.is_absolute():
                path = (base_dir / path).resolve()
    if create:
        path.mkdir(parents=True, exist_ok=True)
    return path


def _infer_format_from_path(path: Path) -> str:
    """Infer data format from file extension.

    Args:
        path: File path

    Returns:
        Format string ('parquet', 'feather', or 'csv')
    """
    suffix = path.suffix.lower()
    if suffix in {".parquet", ".pq"}:
        return "parquet"
    if suffix in {".feather", ".ft"}:
        return "feather"
    return "csv"


def resolve_data_path(
    data_dir: Path,
    model_name: str,
    *,
    data_format: str = "csv",
    path_template: Optional[str] = None,
) -> Path:
    """Resolve the path to a data file.

    Args:
        data_dir: Directory containing data files
        model_name: Model name for path substitution
        data_format: Data format ('csv', 'parquet', 'feather', 'auto')
        path_template: Template string with {model_name} and {ext} placeholders

    Returns:
        Resolved path to data file
    """
    fmt = str(data_format or "csv").strip().lower()
    template = path_template or "{model_name}.{ext}"

    if fmt == "auto":
        candidates = [
            data_dir / template.format(model_name=model_name, ext="parquet"),
            data_dir / template.format(model_name=model_name, ext="feather"),
            data_dir / template.format(model_name=model_name, ext="csv"),
        ]
        for cand in candidates:
            if cand.exists():
                return cand.resolve()
        return candidates[-1].resolve()

    ext = "csv" if fmt in {"csv"} else fmt
    return (data_dir / template.format(model_name=model_name, ext=ext)).resolve()


# =============================================================================
# Data Loading
# =============================================================================


def load_dataset(
    path: Path,
    *,
    data_format: str = "auto",
    dtype_map: Optional[Dict[str, Any]] = None,
    usecols: Optional[List[str]] = None,
    low_memory: bool = False,
    chunksize: Optional[int] = None,
) -> Union[pd.DataFrame, Iterator[pd.DataFrame]]:
    """Load a dataset from various formats.

    Args:
        path: Path to data file
        data_format: Format ('csv', 'parquet', 'feather', 'auto')
        dtype_map: Column type mapping
        usecols: Optional projected columns to load
        low_memory: Whether to use low memory mode for CSV
        chunksize: Optional chunk size for CSV streaming

    Returns:
        Loaded DataFrame, or a CSV chunk iterator when `chunksize` is set
    """
    fmt = str(data_format or "auto").strip().lower()
    if fmt == "auto":
        fmt = _infer_format_from_path(path)
    selected_cols: Optional[List[str]] = None
    if usecols is not None:
        selected_cols = []
        seen = set()
        for col in usecols:
            if not isinstance(col, str):
                continue
            key = col.strip()
            if not key or key in seen:
                continue
            seen.add(key)
            selected_cols.append(key)
        if not selected_cols:
            selected_cols = None

    dtype_selected: Optional[Dict[str, Any]] = None
    if dtype_map:
        dtype_selected = dict(dtype_map)
        if selected_cols is not None:
            allowed = set(selected_cols)
            dtype_selected = {
                k: v for k, v in dtype_selected.items() if k in allowed
            }
        if not dtype_selected:
            dtype_selected = None

    if fmt == "parquet":
        if selected_cols is not None:
            df = pd.read_parquet(path, columns=selected_cols)
        else:
            df = pd.read_parquet(path)
    elif fmt == "feather":
        if selected_cols is not None:
            df = pd.read_feather(path, columns=selected_cols)
        else:
            df = pd.read_feather(path)
    elif fmt == "csv":
        read_kwargs: Dict[str, Any] = {"low_memory": low_memory}
        if dtype_selected is not None:
            read_kwargs["dtype"] = dtype_selected
        if selected_cols is not None:
            read_kwargs["usecols"] = selected_cols
        if chunksize is not None:
            return pd.read_csv(path, chunksize=int(chunksize), **read_kwargs)
        else:
            df = pd.read_csv(path, **read_kwargs)
    else:
        raise ValueError(f"Unsupported data_format: {data_format}")

    if dtype_selected:
        for col, dtype in dtype_selected.items():
            if col in df.columns:
                df[col] = df[col].astype(dtype)
    return df


def coerce_dataset_types(raw: pd.DataFrame) -> pd.DataFrame:
    """Coerce dataset types to numeric or object.

    Args:
        raw: Input DataFrame

    Returns:
        DataFrame with coerced types
    """
    data = raw
    for col in data.columns:
        s = data[col]
        if pd.api.types.is_bool_dtype(s):
            data[col] = s.astype(np.int8, copy=False)
            continue
        if pd.api.types.is_numeric_dtype(s):
            s_num = pd.to_numeric(s, errors="coerce").fillna(0)
            data[col] = s_num.astype(np.float32, copy=False)
            continue
        if pd.api.types.is_object_dtype(s) and s.hasnans:
            # Keep string-like columns as-is, but normalize missing markers.
            data[col] = s.where(s.notna(), "<NA>")
    return data


# =============================================================================
# File Fingerprinting
# =============================================================================


def fingerprint_file(path: Path, *, max_bytes: int = 10_485_760) -> Dict[str, Any]:
    """Generate a fingerprint for a file (for cache validation).

    Args:
        path: Path to file
        max_bytes: Maximum bytes to hash (default 10MB)

    Returns:
        Dictionary with path, size, mtime, and partial SHA256 hash
    """
    path = Path(path)
    stat = path.stat()
    h = hashlib.sha256()
    remaining = int(max_bytes)

    with path.open("rb") as fh:
        while remaining > 0:
            chunk = fh.read(min(1024 * 1024, remaining))
            if not chunk:
                break
            h.update(chunk)
            remaining -= len(chunk)

    return {
        "path": str(path),
        "size": int(stat.st_size),
        "mtime": int(stat.st_mtime),
        "sha256_prefix": h.hexdigest(),
        "max_bytes": int(max_bytes),
    }



"""
Insurance Pricing Model Training Frontend
A Gradio-based web interface for configuring and running insurance pricing models.
"""

import os
import platform
import subprocess
from ins_pricing.frontend.workflows import (
    run_compare_ft_embed,
    run_double_lift_from_file,
    run_plot_direct,
    run_plot_embed,
    run_predict_ft_embed,
    run_pre_oneway,
)
from ins_pricing.frontend.ft_workflow import FTWorkflowHelper
from ins_pricing.frontend.runner import TaskRunner
from ins_pricing.frontend.config_builder import ConfigBuilder
import json
import tempfile
import inspect
import importlib.util
from pathlib import Path
from typing import Optional, Dict, Any, Callable, Iterable, Generator

def _ensure_repo_root() -> None:
    if __package__ not in {None, ""}:
        return
    if importlib.util.find_spec("ins_pricing") is not None:
        return
    bootstrap_path = Path(__file__).resolve().parents[1] / "cli" / "utils" / "bootstrap.py"
    spec = importlib.util.spec_from_file_location("ins_pricing.cli.utils.bootstrap", bootstrap_path)
    if spec is None or spec.loader is None:
        return
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    module.ensure_repo_root()


_ensure_repo_root()

os.environ.setdefault("GRADIO_ANALYTICS_ENABLED", "False")
os.environ.setdefault("GRADIO_TELEMETRY_ENABLED", "False")
os.environ.setdefault("GRADIO_CHECK_VERSION", "False")
os.environ.setdefault("GRADIO_VERSION_CHECK", "False")




class FrontendDependencyError(RuntimeError):
    pass


def _check_frontend_deps() -> None:
    """Fail fast with a clear message if frontend deps are incompatible."""
    try:
        import gradio  # noqa: F401
    except Exception as exc:
        raise FrontendDependencyError(f"Failed to import gradio: {exc}")

    try:
        import huggingface_hub as hf  # noqa: F401
    except Exception as exc:
        raise FrontendDependencyError(
            f"Failed to import huggingface_hub: {exc}. "
            "Pin version with `pip install 'huggingface_hub<0.24'`."
        )

    if not hasattr(hf, 'HfFolder'):
        raise FrontendDependencyError(
            'Incompatible huggingface_hub detected: missing HfFolder. '
            'Please install `huggingface_hub<0.24`.'
        )


class PricingApp:
    """Main application class for the insurance pricing model tasks interface."""

    def __init__(self):
        self.config_builder = ConfigBuilder()
        self.runner = TaskRunner()
        self.ft_workflow = FTWorkflowHelper()
        self.working_dir: Path = Path.cwd().resolve()
        self.current_config = {}
        self.current_step1_config = None
        self.current_step2_config_paths: Dict[str, str] = {}
        self.current_config_path: Optional[Path] = None
        self.current_config_dir: Optional[Path] = None
        self.current_workflow_config: Dict[str, Any] = {}
        self.current_workflow_config_path: Optional[Path] = None
        self.current_workflow_config_dir: Optional[Path] = None

    def load_json_config(self, file_path) -> tuple[str, Dict[str, Any], str]:
        """Load configuration from uploaded JSON file."""
        if not file_path:
            return "No file uploaded", {}, ""

        try:
            path = Path(file_path).resolve()
            with open(path, 'r', encoding='utf-8') as f:
                config = json.load(f)
            self.current_config = config
            self.current_config_path = path
            self.current_config_dir = path.parent
            config_json = json.dumps(config, indent=2, ensure_ascii=False)
            return f"Configuration loaded successfully from {path.name}", config, config_json
        except Exception as e:
            return f"Error loading config: {str(e)}", {}, ""

    def set_working_dir(self, working_dir: str) -> tuple[str, str]:
        """Set working directory used for relative paths and generated configs."""
        try:
            raw = str(working_dir or "").strip()
            if not raw:
                resolved = Path.cwd().resolve()
            else:
                candidate = Path(raw).expanduser()
                if not candidate.is_absolute():
                    candidate = (self.working_dir / candidate).resolve()
                else:
                    candidate = candidate.resolve()
                if not candidate.exists():
                    return f"Working directory does not exist: {candidate}", str(self.working_dir)
                if not candidate.is_dir():
                    return f"Working directory is not a folder: {candidate}", str(self.working_dir)
                resolved = candidate

            self.working_dir = resolved
            return f"Working directory set to: {resolved}", str(resolved)
        except Exception as e:
            return f"Error setting working directory: {str(e)}", str(self.working_dir)

    def list_directory_candidates(
        self,
        root_dir: str,
        *,
        max_depth: int = 2,
        max_items: int = 300,
    ) -> tuple[str, list[str], str]:
        """List directories for manual working-directory selection."""
        try:
            raw_root = str(root_dir or "").strip()
            candidate = Path(raw_root).expanduser() if raw_root else self.working_dir
            if not candidate.is_absolute():
                candidate = (self.working_dir / candidate).resolve()
            else:
                candidate = candidate.resolve()

            if not candidate.exists():
                fallback = str(self.working_dir)
                return f"Browse root does not exist: {candidate}", [fallback], fallback
            if not candidate.is_dir():
                fallback = str(self.working_dir)
                return f"Browse root is not a folder: {candidate}", [fallback], fallback

            depth_limit = max(0, int(max_depth))
            item_limit = max(1, int(max_items))
            root = candidate
            root_parts = len(root.parts)
            dirs: list[str] = []

            for current, child_dirs, _ in os.walk(root):
                current_path = Path(current).resolve()
                rel_depth = len(current_path.parts) - root_parts
                dirs.append(str(current_path))

                if rel_depth >= depth_limit:
                    child_dirs[:] = []

                if len(dirs) >= item_limit:
                    break

            dirs = sorted(dict.fromkeys(dirs))
            selected = str(self.working_dir) if str(self.working_dir) in dirs else str(root)
            truncated = len(dirs) >= item_limit
            status = f"Found {len(dirs)} folders under: {root}"
            if truncated:
                status += f" (limited to first {item_limit})"
            return status, dirs, selected
        except Exception as e:
            fallback = str(self.working_dir)
            return f"Error listing folders: {str(e)}", [fallback], fallback

    def _resolve_user_path(self, value: str, *, base_dir: Optional[Path] = None) -> Path:
        raw = str(value or "").strip()
        if not raw:
            raise ValueError("Path value is empty.")
        path = Path(raw).expanduser()
        root = (base_dir or self.working_dir or Path.cwd()).resolve()
        if not path.is_absolute():
            path = (root / path).resolve()
        else:
            path = path.resolve()
        return path

    def _default_base_dir(self, preferred: Optional[Path] = None) -> Path:
        return (preferred or self.working_dir or Path.cwd()).resolve()

    @staticmethod
    def _parse_json_dict(raw_json: str, field_name: str) -> Dict[str, Any]:
        text = str(raw_json or "").strip()
        if not text:
            return {}
        obj = json.loads(text)
        if not isinstance(obj, dict):
            raise ValueError(f"{field_name} must be a JSON object.")
        return obj

    def build_config_from_ui(
        self,
        data_dir: str,
        model_list: str,
        model_categories: str,
        target: str,
        weight: str,
        feature_list: str,
        categorical_features: str,
        task_type: str,
        prop_test: float,
        holdout_ratio: float,
        val_ratio: float,
        split_strategy: str,
        rand_seed: int,
        epochs: int,
        output_dir: str,
        use_gpu: bool,
        model_keys: str,
        max_evals: int,
        xgb_max_depth_max: int,
        xgb_n_estimators_max: int,
        xgb_gpu_id: int,
        xgb_cleanup_per_fold: bool,
        xgb_cleanup_synchronize: bool,
        xgb_use_dmatrix: bool,
        xgb_chunk_size: int,
        xgb_search_space_json: str,
        resn_search_space_json: str,
        ft_search_space_json: str,
        ft_unsupervised_search_space_json: str,
        ft_cleanup_per_fold: bool,
        ft_cleanup_synchronize: bool,
        resn_cleanup_per_fold: bool,
        resn_cleanup_synchronize: bool,
        resn_use_lazy_dataset: bool,
        resn_predict_batch_size: int,
        gnn_cleanup_per_fold: bool,
        gnn_cleanup_synchronize: bool,
        optuna_cleanup_synchronize: bool,
    ) -> tuple[str, str]:
        """Build configuration from UI parameters."""
        try:
            # Parse comma-separated lists
            model_list = [x.strip()
                          for x in model_list.split(',') if x.strip()]
            model_categories = [x.strip()
                                for x in model_categories.split(',') if x.strip()]
            feature_list = [x.strip()
                            for x in feature_list.split(',') if x.strip()]
            categorical_features = [
                x.strip() for x in categorical_features.split(',') if x.strip()]
            model_keys = [x.strip()
                          for x in model_keys.split(',') if x.strip()]
            parsed_xgb_chunk_size: Optional[int] = None
            try:
                chunk_size_val = int(xgb_chunk_size)
            except (TypeError, ValueError):
                chunk_size_val = 0
            if chunk_size_val > 0:
                parsed_xgb_chunk_size = chunk_size_val
            parsed_resn_predict_batch_size: Optional[int] = None
            try:
                resn_pred_bs_val = int(resn_predict_batch_size)
            except (TypeError, ValueError):
                resn_pred_bs_val = 0
            if resn_pred_bs_val > 0:
                parsed_resn_predict_batch_size = resn_pred_bs_val
            xgb_search_space = self._parse_json_dict(
                xgb_search_space_json,
                "xgb_search_space_json",
            )
            resn_search_space = self._parse_json_dict(
                resn_search_space_json,
                "resn_search_space_json",
            )
            ft_search_space = self._parse_json_dict(
                ft_search_space_json,
                "ft_search_space_json",
            )
            ft_unsupervised_search_space = self._parse_json_dict(
                ft_unsupervised_search_space_json,
                "ft_unsupervised_search_space_json",
            )

            config = self.config_builder.build_config(
                data_dir=data_dir,
                model_list=model_list,
                model_categories=model_categories,
                target=target,
                weight=weight,
                feature_list=feature_list,
                categorical_features=categorical_features,
                task_type=task_type,
                prop_test=prop_test,
                holdout_ratio=holdout_ratio,
                val_ratio=val_ratio,
                split_strategy=split_strategy,
                rand_seed=rand_seed,
                epochs=epochs,
                output_dir=output_dir,
                use_gpu=use_gpu,
                model_keys=model_keys,
                max_evals=max_evals,
                xgb_max_depth_max=xgb_max_depth_max,
                xgb_n_estimators_max=xgb_n_estimators_max,
                xgb_gpu_id=xgb_gpu_id,
                xgb_cleanup_per_fold=xgb_cleanup_per_fold,
                xgb_cleanup_synchronize=xgb_cleanup_synchronize,
                xgb_use_dmatrix=xgb_use_dmatrix,
                xgb_chunk_size=parsed_xgb_chunk_size,
                xgb_search_space=xgb_search_space,
                resn_search_space=resn_search_space,
                ft_search_space=ft_search_space,
                ft_unsupervised_search_space=ft_unsupervised_search_space,
                ft_cleanup_per_fold=ft_cleanup_per_fold,
                ft_cleanup_synchronize=ft_cleanup_synchronize,
                resn_cleanup_per_fold=resn_cleanup_per_fold,
                resn_cleanup_synchronize=resn_cleanup_synchronize,
                resn_use_lazy_dataset=resn_use_lazy_dataset,
                resn_predict_batch_size=parsed_resn_predict_batch_size,
                gnn_cleanup_per_fold=gnn_cleanup_per_fold,
                gnn_cleanup_synchronize=gnn_cleanup_synchronize,
                optuna_cleanup_synchronize=optuna_cleanup_synchronize,
            )

            is_valid, msg = self.config_builder.validate_config(config)
            if not is_valid:
                return f"Validation failed: {msg}", ""

            self.current_config = config
            self.current_config_path = None
            self.current_config_dir = self.working_dir
            config_json = json.dumps(config, indent=2, ensure_ascii=False)
            return "Configuration built successfully", config_json

        except Exception as e:
            return f"Error building config: {str(e)}", ""

    def save_config(self, config_json: str, filename: str) -> str:
        """Save current configuration to file."""
        if not config_json:
            return "No configuration to save"

        try:
            config_path = self._resolve_user_path(filename, base_dir=self._default_base_dir(self.current_config_dir))
            config_path.parent.mkdir(parents=True, exist_ok=True)
            with open(config_path, 'w', encoding='utf-8') as f:
                json.dump(json.loads(config_json), f,
                          indent=2, ensure_ascii=False)
            return f"Configuration saved to {config_path}"
        except Exception as e:
            return f"Error saving config: {str(e)}"

    def load_workflow_config(self, file_path) -> tuple[str, str]:
        """Load workflow configuration (plot/predict/compare/pre-oneway)."""
        if not file_path:
            return "No file uploaded", ""

        try:
            path = Path(file_path).resolve()
            with open(path, 'r', encoding='utf-8') as f:
                config = json.load(f)
            self.current_workflow_config = config
            self.current_workflow_config_path = path
            self.current_workflow_config_dir = path.parent
            config_json = json.dumps(config, indent=2, ensure_ascii=False)
            return f"Workflow config loaded from {path.name}", config_json
        except Exception as e:
            return f"Error loading workflow config: {str(e)}", ""

    @staticmethod
    def _to_csv(value: Any) -> str:
        if value is None:
            return ""
        if isinstance(value, str):
            return value
        if isinstance(value, (list, tuple, set)):
            return ", ".join(str(x).strip() for x in value if str(x).strip())
        return str(value)

    @staticmethod
    def _resolve_path_value(
        base_dir: Path,
        value: Any,
        field_name: str,
        *,
        required: bool = True,
    ) -> Optional[str]:
        raw = str(value or "").strip()
        if not raw:
            if required:
                raise ValueError(f"{field_name} is required.")
            return None
        path = Path(raw)
        if not path.is_absolute():
            path = (base_dir / path).resolve()
        else:
            path = path.resolve()
        return str(path)

    def _run_workflow_from_config(self, config: Dict[str, Any], base_dir: Path) -> str:
        workflow_cfg = config.get("workflow", config)
        if not isinstance(workflow_cfg, dict):
            raise ValueError("workflow must be a JSON object.")

        mode = str(workflow_cfg.get("mode", "")).strip().lower()
        if not mode:
            raise ValueError(
                "workflow.mode is required. Supported modes: "
                "pre_oneway, plot_direct, plot_embed, predict_ft_embed, compare_xgb, compare_resn, compare, double_lift."
            )

        if mode in {"pre_oneway", "pre-oneway", "oneway_pre"}:
            holdout_ratio_raw = workflow_cfg.get("holdout_ratio", 0.25)
            holdout_ratio = None if holdout_ratio_raw is None else float(holdout_ratio_raw)
            output_dir = self._resolve_path_value(
                base_dir, workflow_cfg.get("output_dir"), "output_dir", required=False
            )
            return run_pre_oneway(
                data_path=self._resolve_path_value(base_dir, workflow_cfg.get("data_path"), "data_path"),
                model_name=str(workflow_cfg.get("model_name", "")).strip(),
                target_col=str(workflow_cfg.get("target_col", "")).strip(),
                weight_col=str(workflow_cfg.get("weight_col", "")).strip(),
                feature_list=self._to_csv(workflow_cfg.get("feature_list", "")),
                categorical_features=self._to_csv(workflow_cfg.get("categorical_features", "")),
                n_bins=int(workflow_cfg.get("n_bins", 10)),
                holdout_ratio=holdout_ratio,
                rand_seed=int(workflow_cfg.get("rand_seed", 13)),
                output_dir=output_dir,
            )

        if mode == "plot_direct":
            cfg_path = workflow_cfg.get("cfg_path", workflow_cfg.get("plot_cfg_path"))
            return run_plot_direct(
                cfg_path=self._resolve_path_value(base_dir, cfg_path, "cfg_path"),
                xgb_cfg_path=self._resolve_path_value(base_dir, workflow_cfg.get("xgb_cfg_path"), "xgb_cfg_path"),
                resn_cfg_path=self._resolve_path_value(base_dir, workflow_cfg.get("resn_cfg_path"), "resn_cfg_path"),
            )

        if mode == "plot_embed":
            cfg_path = workflow_cfg.get("cfg_path", workflow_cfg.get("plot_cfg_path"))
            return run_plot_embed(
                cfg_path=self._resolve_path_value(base_dir, cfg_path, "cfg_path"),
                xgb_cfg_path=self._resolve_path_value(base_dir, workflow_cfg.get("xgb_cfg_path"), "xgb_cfg_path"),
                resn_cfg_path=self._resolve_path_value(base_dir, workflow_cfg.get("resn_cfg_path"), "resn_cfg_path"),
                ft_cfg_path=self._resolve_path_value(base_dir, workflow_cfg.get("ft_cfg_path"), "ft_cfg_path"),
                use_runtime_ft_embedding=bool(workflow_cfg.get("use_runtime_ft_embedding", False)),
            )

        if mode in {"predict_ft_embed", "predict"}:
            xgb_cfg_path = self._resolve_path_value(
                base_dir, workflow_cfg.get("xgb_cfg_path"), "xgb_cfg_path", required=False
            )
            resn_cfg_path = self._resolve_path_value(
                base_dir, workflow_cfg.get("resn_cfg_path"), "resn_cfg_path", required=False
            )
            return run_predict_ft_embed(
                ft_cfg_path=self._resolve_path_value(base_dir, workflow_cfg.get("ft_cfg_path"), "ft_cfg_path"),
                xgb_cfg_path=xgb_cfg_path,
                resn_cfg_path=resn_cfg_path,
                input_path=self._resolve_path_value(base_dir, workflow_cfg.get("input_path"), "input_path"),
                output_path=self._resolve_path_value(base_dir, workflow_cfg.get("output_path"), "output_path"),
                model_name=(str(workflow_cfg.get("model_name", "")).strip() or None),
                model_keys=self._to_csv(workflow_cfg.get("model_keys", "")),
            )

        if mode in {"compare_xgb", "compare_resn", "compare"}:
            if mode == "compare_xgb":
                model_key = "xgb"
            elif mode == "compare_resn":
                model_key = "resn"
            else:
                model_key = str(workflow_cfg.get("model_key", "xgb")).strip().lower()
            if model_key not in {"xgb", "resn"}:
                raise ValueError("compare mode only supports model_key in {'xgb', 'resn'}.")

            n_bins_override_raw = workflow_cfg.get("n_bins_override", 10)
            n_bins_override = None if n_bins_override_raw is None else int(n_bins_override_raw)
            return run_compare_ft_embed(
                direct_cfg_path=self._resolve_path_value(base_dir, workflow_cfg.get("direct_cfg_path"), "direct_cfg_path"),
                ft_cfg_path=self._resolve_path_value(base_dir, workflow_cfg.get("ft_cfg_path"), "ft_cfg_path"),
                ft_embed_cfg_path=self._resolve_path_value(base_dir, workflow_cfg.get("ft_embed_cfg_path"), "ft_embed_cfg_path"),
                model_key=model_key,
                label_direct=str(workflow_cfg.get("label_direct", "Direct")).strip(),
                label_ft=str(workflow_cfg.get("label_ft", "FT")).strip(),
                use_runtime_ft_embedding=bool(workflow_cfg.get("use_runtime_ft_embedding", False)),
                n_bins_override=n_bins_override,
            )

        if mode in {"double_lift", "double-lift"}:
            holdout_ratio_raw = workflow_cfg.get("holdout_ratio", 0.0)
            holdout_ratio = None if holdout_ratio_raw is None else float(holdout_ratio_raw)
            output_path = self._resolve_path_value(
                base_dir, workflow_cfg.get("output_path"), "output_path", required=False
            )
            return run_double_lift_from_file(
                data_path=self._resolve_path_value(base_dir, workflow_cfg.get("data_path"), "data_path"),
                pred_col_1=str(workflow_cfg.get("pred_col_1", workflow_cfg.get("pred_col1", ""))).strip(),
                pred_col_2=str(workflow_cfg.get("pred_col_2", workflow_cfg.get("pred_col2", ""))).strip(),
                target_col=str(workflow_cfg.get("target_col", workflow_cfg.get("target", ""))).strip(),
                weight_col=str(workflow_cfg.get("weight_col", workflow_cfg.get("weight", "weights"))).strip(),
                n_bins=int(workflow_cfg.get("n_bins", 10)),
                label1=str(workflow_cfg.get("label1", "")).strip() or None,
                label2=str(workflow_cfg.get("label2", "")).strip() or None,
                pred1_weighted=bool(workflow_cfg.get("pred1_weighted", False)),
                pred2_weighted=bool(workflow_cfg.get("pred2_weighted", False)),
                actual_weighted=bool(workflow_cfg.get("actual_weighted", False)),
                holdout_ratio=holdout_ratio,
                split_strategy=str(workflow_cfg.get("split_strategy", "random")).strip(),
                split_group_col=str(workflow_cfg.get("split_group_col", "")).strip() or None,
                split_time_col=str(workflow_cfg.get("split_time_col", "")).strip() or None,
                split_time_ascending=bool(workflow_cfg.get("split_time_ascending", True)),
                rand_seed=int(workflow_cfg.get("rand_seed", 13)),
                output_path=output_path,
            )

        raise ValueError(
            f"Unsupported workflow mode: {mode}. "
            "Supported: pre_oneway, plot_direct, plot_embed, predict_ft_embed, compare_xgb, compare_resn, compare, double_lift."
        )

    def run_workflow_config_ui(self, workflow_config_json: str) -> Generator[tuple[str, str], None, None]:
        """Run plotting/prediction/compare/pre-oneway from workflow JSON config."""
        try:
            if workflow_config_json:
                workflow_config = json.loads(workflow_config_json)
                base_dir = self._default_base_dir(self.current_workflow_config_dir)
                self.current_workflow_config = workflow_config
            elif self.current_workflow_config:
                workflow_config = self.current_workflow_config
                base_dir = self._default_base_dir(self.current_workflow_config_dir)
            else:
                yield "No workflow configuration provided", ""
                return

            mode = str(workflow_config.get("workflow", workflow_config).get("mode", "unknown")).strip().lower()
            log_generator = self.runner.run_callable(
                self._run_workflow_from_config,
                workflow_config,
                base_dir,
            )

            full_log = ""
            for log_line in log_generator:
                full_log += log_line + "\n"
                yield f"Workflow [{mode}] in progress...", full_log

            yield f"Workflow [{mode}] completed!", full_log

        except Exception as e:
            error_msg = f"Workflow config execution error: {str(e)}"
            yield error_msg, error_msg

    def run_training(self, config_json: str) -> Generator[tuple[str, str], None, None]:
        """
        Run task (training, explain, plotting, etc.) with the current configuration.

        The task type is automatically detected from config.runner.mode.
        Supported modes: entry (training), explain, incremental, watchdog, etc.
        """
        temp_config_path: Optional[Path] = None
        try:
            if config_json:
                config = json.loads(config_json)
                task_mode = config.get('runner', {}).get('mode', 'entry')
                base_dir = self._default_base_dir(self.current_config_dir)
                fd, temp_path = tempfile.mkstemp(prefix="temp_config_", suffix=".json", dir=base_dir)
                temp_config_path = Path(temp_path)
                with os.fdopen(fd, 'w', encoding='utf-8') as f:
                    json.dump(config, f, indent=2)
                config_path = temp_config_path
            elif self.current_config_path and self.current_config_path.exists():
                config_path = self.current_config_path
                config = json.loads(config_path.read_text(encoding="utf-8"))
                task_mode = config.get('runner', {}).get('mode', 'entry')
            elif self.current_config:
                config = self.current_config
                task_mode = config.get('runner', {}).get('mode', 'entry')
                base_dir = self._default_base_dir(self.current_config_dir)
                fd, temp_path = tempfile.mkstemp(prefix="temp_config_", suffix=".json", dir=base_dir)
                temp_config_path = Path(temp_path)
                with os.fdopen(fd, 'w', encoding='utf-8') as f:
                    json.dump(config, f, indent=2)
                config_path = temp_config_path
            else:
                yield "No configuration provided", ""
                return

            log_generator = self.runner.run_task(str(config_path))

            # Collect logs
            full_log = ""
            for log_line in log_generator:
                full_log += log_line + "\n"
                yield f"Task [{task_mode}] in progress...", full_log

            yield f"Task [{task_mode}] completed!", full_log

        except Exception as e:
            error_msg = f"Error during task execution: {str(e)}"
            yield error_msg, error_msg
        finally:
            if temp_config_path is not None:
                try:
                    temp_config_path.unlink(missing_ok=True)
                except Exception:
                    pass

    def prepare_ft_step1(self, config_json: str, use_ddp: bool, nproc: int) -> tuple[str, str]:
        """Prepare FT Step 1 configuration."""
        if not config_json:
            return "No configuration provided", ""

        try:
            config = json.loads(config_json)
            step1_config = self.ft_workflow.prepare_step1_config(
                base_config=config,
                use_ddp=use_ddp,
                nproc_per_node=int(nproc)
            )

            # Save to temp file
            base_dir = self._default_base_dir(self.current_config_dir)
            temp_path = (base_dir / "temp_ft_step1_config.json").resolve()
            with open(temp_path, 'w', encoding='utf-8') as f:
                json.dump(step1_config, f, indent=2)

            self.current_step1_config = str(temp_path)
            step1_json = json.dumps(step1_config, indent=2, ensure_ascii=False)

            return "Step 1 config prepared. Click 'Run Step 1' to train FT embeddings.", step1_json

        except Exception as e:
            return f"Error preparing Step 1 config: {str(e)}", ""

    def prepare_ft_step2(
        self,
        step1_config_path: str,
        target_models: str,
        augmented_data_dir: str,
        xgb_overrides_json: str,
        resn_overrides_json: str,
    ) -> tuple[str, str, str]:
        """Prepare FT Step 2 configurations."""
        if not step1_config_path:
            return "Step 1 config not found. Run Step 1 first.", "", ""

        try:
            step1_path = Path(step1_config_path).expanduser()
            if not step1_path.is_absolute():
                candidate_dirs: Iterable[Path] = (
                    self.current_config_dir,
                    self.working_dir,
                    Path.cwd(),
                )
                resolved_candidate = None
                for root in candidate_dirs:
                    if root is None:
                        continue
                    candidate = (Path(root).resolve() / step1_path).resolve()
                    if candidate.exists():
                        resolved_candidate = candidate
                        break
                if resolved_candidate is None:
                    return "Step 1 config not found. Run Step 1 first.", "", ""
                step1_path = resolved_candidate
            else:
                step1_path = step1_path.resolve()
                if not step1_path.exists():
                    return "Step 1 config not found. Run Step 1 first.", "", ""

            models = [m.strip() for m in target_models.split(',') if m.strip()]
            data_dir_value = str(augmented_data_dir or "").strip() or "./DataFTUnsupervised"
            xgb_overrides = self._parse_json_dict(
                xgb_overrides_json,
                "xgb_overrides_json",
            )
            resn_overrides = self._parse_json_dict(
                resn_overrides_json,
                "resn_overrides_json",
            )
            xgb_cfg, resn_cfg = self.ft_workflow.generate_step2_configs(
                step1_config_path=str(step1_path),
                target_models=models,
                augmented_data_dir=data_dir_value,
                xgb_overrides=xgb_overrides,
                resn_overrides=resn_overrides,
            )
            save_dir = step1_path.parent
            saved_paths: Dict[str, str] = {}

            if xgb_cfg:
                xgb_cfg_path = save_dir / "config_xgb_from_ft_unsupervised.json"
                xgb_cfg_path.write_text(
                    json.dumps(xgb_cfg, indent=2, ensure_ascii=False),
                    encoding="utf-8",
                )
                saved_paths["xgb"] = str(xgb_cfg_path)

            if resn_cfg:
                resn_cfg_path = save_dir / "config_resn_from_ft_unsupervised.json"
                resn_cfg_path.write_text(
                    json.dumps(resn_cfg, indent=2, ensure_ascii=False),
                    encoding="utf-8",
                )
                saved_paths["resn"] = str(resn_cfg_path)

            self.current_step2_config_paths = saved_paths

            status_lines = [
                f"Step 2 configs prepared for: {', '.join(models)}",
                f"Augmented data dir: {data_dir_value}",
            ]
            if "xgb" in saved_paths:
                status_lines.append(f"Saved XGB config: {saved_paths['xgb']}")
            if "resn" in saved_paths:
                status_lines.append(f"Saved ResN config: {saved_paths['resn']}")
            status_msg = "\n".join(status_lines)
            xgb_json = json.dumps(
                xgb_cfg, indent=2, ensure_ascii=False) if xgb_cfg else ""
            resn_json = json.dumps(
                resn_cfg, indent=2, ensure_ascii=False) if resn_cfg else ""

            return status_msg, xgb_json, resn_json

        except FileNotFoundError as e:
            return f"Error: {str(e)}\n\nMake sure Step 1 completed successfully.", "", ""
        except Exception as e:
            return f"Error preparing Step 2 configs: {str(e)}", "", ""

    def open_results_folder(self, config_json: str) -> str:
        """Open the results folder in file explorer."""
        try:
            if config_json:
                config = json.loads(config_json)
                output_dir = config.get('output_dir', './Results')
                out = Path(str(output_dir))
                if out.is_absolute():
                    results_path = out.resolve()
                else:
                    results_path = (self._default_base_dir(self.current_config_dir) / out).resolve()
            elif self.current_config_path and self.current_config_path.exists():
                config = json.loads(
                    self.current_config_path.read_text(encoding="utf-8"))
                output_dir = config.get('output_dir', './Results')
                results_path = (
                    self.current_config_path.parent / output_dir).resolve()
            elif self.current_config:
                output_dir = self.current_config.get('output_dir', './Results')
                out = Path(str(output_dir))
                if out.is_absolute():
                    results_path = out.resolve()
                else:
                    results_path = (self._default_base_dir(self.current_config_dir) / out).resolve()
            else:
                return "No configuration loaded"

            if not results_path.exists():
                return f"Results folder does not exist yet: {results_path}"

            # Open folder based on OS
            system = platform.system()
            if system == "Windows":
                os.startfile(results_path)
            elif system == "Darwin":  # macOS
                subprocess.run(["open", str(results_path)])
            else:  # Linux
                subprocess.run(["xdg-open", str(results_path)])

            return f"Opened folder: {results_path}"

        except Exception as e:
            return f"Error opening folder: {str(e)}"

    def _run_workflow(self, label: str, func: Callable, *args, **kwargs):
        """Run a workflow function and stream logs."""
        try:
            log_generator = self.runner.run_callable(func, *args, **kwargs)
            full_log = ""
            for log_line in log_generator:
                full_log += log_line + "\n"
                yield f"{label} in progress...", full_log
            yield f"{label} completed!", full_log
        except Exception as e:
            error_msg = f"{label} error: {str(e)}"
            yield error_msg, error_msg

    def run_pre_oneway_ui(
        self,
        data_path: str,
        model_name: str,
        target_col: str,
        weight_col: str,
        feature_list: str,
        categorical_features: str,
        n_bins: int,
        holdout_ratio: float,
        rand_seed: int,
        output_dir: str,
    ):
        yield from self._run_workflow(
            "Pre-Oneway Plot",
            run_pre_oneway,
            data_path=data_path,
            model_name=model_name,
            target_col=target_col,
            weight_col=weight_col,
            feature_list=feature_list,
            categorical_features=categorical_features,
            n_bins=n_bins,
            holdout_ratio=holdout_ratio,
            rand_seed=rand_seed,
            output_dir=output_dir or None,
        )

    def run_plot_direct_ui(self, cfg_path: str, xgb_cfg_path: str, resn_cfg_path: str):
        yield from self._run_workflow(
            "Direct Plot",
            run_plot_direct,
            cfg_path=cfg_path,
            xgb_cfg_path=xgb_cfg_path,
            resn_cfg_path=resn_cfg_path,
        )

    def run_plot_embed_ui(
        self,
        cfg_path: str,
        xgb_cfg_path: str,
        resn_cfg_path: str,
        ft_cfg_path: str,
        use_runtime_ft_embedding: bool,
    ):
        yield from self._run_workflow(
            "Embed Plot",
            run_plot_embed,
            cfg_path=cfg_path,
            xgb_cfg_path=xgb_cfg_path,
            resn_cfg_path=resn_cfg_path,
            ft_cfg_path=ft_cfg_path,
            use_runtime_ft_embedding=use_runtime_ft_embedding,
        )

    def run_predict_ui(
        self,
        ft_cfg_path: str,
        xgb_cfg_path: str,
        resn_cfg_path: str,
        input_path: str,
        output_path: str,
        model_name: str,
        model_keys: str,
    ):
        yield from self._run_workflow(
            "Prediction",
            run_predict_ft_embed,
            ft_cfg_path=ft_cfg_path,
            xgb_cfg_path=xgb_cfg_path or None,
            resn_cfg_path=resn_cfg_path or None,
            input_path=input_path,
            output_path=output_path,
            model_name=model_name or None,
            model_keys=model_keys,
        )

    def run_compare_ui(
        self,
        model_key: str,
        direct_cfg_path: str,
        ft_cfg_path: str,
        ft_embed_cfg_path: str,
        label_direct: str,
        label_ft: str,
        use_runtime_ft_embedding: bool,
        n_bins_override: int,
    ):
        model_key_norm = str(model_key or "").strip().lower()
        if model_key_norm not in {"xgb", "resn"}:
            raise ValueError("model_key must be one of: xgb, resn.")
        label = "Compare XGB" if model_key_norm == "xgb" else "Compare ResNet"
        yield from self._run_compare_ui(
            model_key=model_key_norm,
            label=label,
            direct_cfg_path=direct_cfg_path,
            ft_cfg_path=ft_cfg_path,
            ft_embed_cfg_path=ft_embed_cfg_path,
            label_direct=label_direct,
            label_ft=label_ft,
            use_runtime_ft_embedding=use_runtime_ft_embedding,
            n_bins_override=n_bins_override,
        )

    def _run_compare_ui(
        self,
        *,
        model_key: str,
        label: str,
        direct_cfg_path: str,
        ft_cfg_path: str,
        ft_embed_cfg_path: str,
        label_direct: str,
        label_ft: str,
        use_runtime_ft_embedding: bool,
        n_bins_override: int,
    ):
        yield from self._run_workflow(
            label,
            run_compare_ft_embed,
            direct_cfg_path=direct_cfg_path,
            ft_cfg_path=ft_cfg_path,
            ft_embed_cfg_path=ft_embed_cfg_path,
            model_key=model_key,
            label_direct=label_direct,
            label_ft=label_ft,
            use_runtime_ft_embedding=use_runtime_ft_embedding,
            n_bins_override=n_bins_override,
        )

    def run_double_lift_ui(
        self,
        data_path: str,
        pred_col_1: str,
        pred_col_2: str,
        target_col: str,
        weight_col: str,
        n_bins: int,
        label1: str,
        label2: str,
        pred1_weighted: bool,
        pred2_weighted: bool,
        actual_weighted: bool,
        holdout_ratio: float,
        split_strategy: str,
        split_group_col: str,
        split_time_col: str,
        split_time_ascending: bool,
        rand_seed: int,
        output_path: str,
    ):
        yield from self._run_workflow(
            "Double Lift",
            run_double_lift_from_file,
            data_path=data_path,
            pred_col_1=pred_col_1,
            pred_col_2=pred_col_2,
            target_col=target_col,
            weight_col=weight_col,
            n_bins=n_bins,
            label1=label1 or None,
            label2=label2 or None,
            pred1_weighted=pred1_weighted,
            pred2_weighted=pred2_weighted,
            actual_weighted=actual_weighted,
            holdout_ratio=holdout_ratio,
            split_strategy=split_strategy,
            split_group_col=split_group_col or None,
            split_time_col=split_time_col or None,
            split_time_ascending=split_time_ascending,
            rand_seed=rand_seed,
            output_path=output_path or None,
        )


def create_ui():
    """Create the Gradio interface."""
    _check_frontend_deps()
    import gradio as gr

    app = PricingApp()
    xgb_search_space_template = json.dumps(
        app.config_builder._default_xgb_search_space(),
        indent=2,
        ensure_ascii=False,
    )
    resn_search_space_template = json.dumps(
        app.config_builder._default_resn_search_space(),
        indent=2,
        ensure_ascii=False,
    )
    ft_search_space_template = json.dumps(
        app.config_builder._default_ft_search_space(),
        indent=2,
        ensure_ascii=False,
    )
    ft_unsupervised_search_space_template = json.dumps(
        app.config_builder._default_ft_unsupervised_search_space(),
        indent=2,
        ensure_ascii=False,
    )
    workflow_template = json.dumps(
        {
            "workflow": {
                "mode": "plot_direct",
                "cfg_path": "config_plot.json",
                "xgb_cfg_path": "config_xgb_direct.json",
                "resn_cfg_path": "config_resn_direct.json",
            }
        },
        indent=2,
        ensure_ascii=False,
    )
    xgb_step2_overrides_template = json.dumps(
        {
            "output_dir": "./ResultsXGBFromFTUnsupervised",
            "optuna_storage": "./ResultsXGBFromFTUnsupervised/optuna/bayesopt.sqlite3",
            "optuna_study_prefix": "pricing_ft_unsup_xgb",
            "loss_name": "mse",
            "build_oht": False,
            "final_refit": False,
            "runner": {
                "model_keys": ["xgb"],
                "nproc_per_node": 1,
                "plot_curves": False,
            },
            "plot_curves": False,
            "plot": {"enable": False},
        },
        indent=2,
        ensure_ascii=False,
    )
    resn_step2_overrides_template = json.dumps(
        {
            "use_resn_ddp": True,
            "output_dir": "./ResultsResNFromFTUnsupervised",
            "optuna_storage": "./ResultsResNFromFTUnsupervised/optuna/bayesopt.sqlite3",
            "optuna_study_prefix": "pricing_ft_unsup_resn_ddp",
            "loss_name": "mse",
            "build_oht": True,
            "runner": {
                "model_keys": ["resn"],
                "nproc_per_node": 2,
                "plot_curves": False,
            },
            "plot_curves": False,
            "plot": {"enable": False},
        },
        indent=2,
        ensure_ascii=False,
    )
    dir_status_init, dir_choices_init, dir_value_init = app.list_directory_candidates(
        str(app.working_dir)
    )

    def _set_working_dir_ui(path_text: str):
        status, resolved = app.set_working_dir(path_text)
        _, choices, selected = app.list_directory_candidates(resolved)
        return (
            status,
            resolved,
            resolved,
            gr.update(choices=choices, value=selected),
        )

    def _refresh_working_dir_choices_ui(root_dir: str):
        status, choices, selected = app.list_directory_candidates(root_dir)
        return status, gr.update(choices=choices, value=selected)

    def _suggest_compare_defaults(model_key: str):
        key = str(model_key or "").strip().lower()
        if key == "resn":
            return (
                "config_resn_direct.json",
                "config_resn_from_ft_unsupervised.json",
                "ResN_raw",
                "ResN_ft_embed",
            )
        return (
            "config_xgb_direct.json",
            "config_xgb_from_ft_unsupervised.json",
            "XGB_raw",
            "XGB_ft_embed",
        )

    layout_css = """
    .gradio-container {
        max-width: 1480px !important;
        margin: 0 auto !important;
    }
    .gradio-container .tabitem {
        padding-top: 8px;
    }
    .gradio-container .gr-form {
        gap: 10px !important;
    }
    .gradio-container .gr-row {
        align-items: stretch;
    }
    .gradio-container .gr-column {
        min-width: 0;
    }
    .gradio-container textarea {
        line-height: 1.35;
    }
    """

    with gr.Blocks(
        title="Insurance Pricing Model Training",
        theme=gr.themes.Soft(),
        css=layout_css,
    ) as demo:
        gr.Markdown(
            """
            # Insurance Pricing Model Training Interface
            Configure and train insurance pricing models with an easy-to-use interface.

            **Two ways to configure:**
            1. **Upload JSON Config**: Upload an existing configuration file
            2. **Manual Configuration**: Fill in the parameters below
            """
        )

        with gr.Row():
            working_dir_input = gr.Textbox(
                label="Working Directory",
                value=str(app.working_dir),
                placeholder="Type a path, or select from the folder list below",
                scale=3,
            )
            set_working_dir_btn = gr.Button(
                "Set Working Directory", variant="secondary", scale=1)

        with gr.Row():
            working_dir_browse_root = gr.Textbox(
                label="Browse Root",
                value=str(app.working_dir),
                placeholder="List folders under this path (depth=2)",
                scale=3,
            )
            refresh_working_dir_btn = gr.Button(
                "Refresh Folder List", variant="secondary", scale=1
            )

        with gr.Row():
            working_dir_picker = gr.Dropdown(
                label="Select Existing Folder",
                choices=dir_choices_init,
                value=dir_value_init,
                scale=3,
            )
            use_selected_working_dir_btn = gr.Button(
                "Use Selected Folder", variant="secondary", scale=1
            )

        working_dir_status = gr.Textbox(
            label="Working Directory Status", value=dir_status_init, interactive=False)

        with gr.Tab("Configuration"):
            with gr.Row(equal_height=True):
                with gr.Column(scale=2):
                    gr.Markdown("### Load Configuration")
                    json_file = gr.File(
                        label="Upload JSON Config File",
                        file_types=[".json"],
                        type="filepath"
                    )
                    load_btn = gr.Button("Load Config", variant="primary")
                    load_status = gr.Textbox(
                        label="Load Status", interactive=False)

                with gr.Column(scale=5):
                    gr.Markdown("### Current Configuration")
                    config_display = gr.JSON(label="Configuration", value={})

            gr.Markdown("---")
            gr.Markdown("### Manual Configuration")

            with gr.Row(equal_height=True):
                with gr.Column(scale=1):
                    gr.Markdown("#### Data Settings")
                    data_dir = gr.Textbox(
                        label="Data Directory", value="./Data")
                    model_list = gr.Textbox(
                        label="Model List (comma-separated)", value="od")
                    model_categories = gr.Textbox(
                        label="Model Categories (comma-separated)", value="bc")
                    target = gr.Textbox(
                        label="Target Column", value="response")
                    weight = gr.Textbox(label="Weight Column", value="weights")

                    gr.Markdown("#### Features")
                    feature_list = gr.Textbox(
                        label="Feature List (comma-separated)",
                        placeholder="feature_1, feature_2, feature_3",
                        lines=4
                    )
                    categorical_features = gr.Textbox(
                        label="Categorical Features (comma-separated)",
                        placeholder="feature_2, feature_3",
                        lines=3
                    )

                with gr.Column(scale=1):
                    gr.Markdown("#### Model Settings")
                    task_type = gr.Dropdown(
                        label="Task Type",
                        choices=["regression", "binary", "multiclass"],
                        value="regression"
                    )
                    split_strategy = gr.Dropdown(
                        label="Split Strategy",
                        choices=["random", "stratified", "time", "group"],
                        value="random"
                    )
                    rand_seed = gr.Number(
                        label="Random Seed", value=13, precision=0)
                    epochs = gr.Number(label="Epochs", value=50, precision=0)
                    prop_test = gr.Slider(
                        label="Test Proportion", minimum=0.1, maximum=0.5, value=0.25, step=0.05)
                    holdout_ratio = gr.Slider(
                        label="Holdout Ratio", minimum=0.1, maximum=0.5, value=0.25, step=0.05)
                    val_ratio = gr.Slider(
                        label="Validation Ratio", minimum=0.1, maximum=0.5, value=0.25, step=0.05)

                    gr.Markdown("#### Training Settings")
                    output_dir = gr.Textbox(
                        label="Output Directory", value="./Results")
                    use_gpu = gr.Checkbox(label="Use GPU", value=True)
                    model_keys = gr.Textbox(
                        label="Model Keys (comma-separated)",
                        value="xgb, resn",
                        placeholder="xgb, resn, ft, gnn"
                    )
                    max_evals = gr.Number(
                        label="Max Evaluations", value=50, precision=0)

                    with gr.Row():
                        with gr.Column(scale=1):
                            gr.Markdown("#### XGBoost Runtime")
                            xgb_max_depth_max = gr.Number(
                                label="XGB Max Depth", value=25, precision=0)
                            xgb_n_estimators_max = gr.Number(
                                label="XGB Max Estimators", value=500, precision=0)
                            xgb_gpu_id = gr.Number(
                                label="XGB GPU ID", value=0, precision=0)
                            xgb_use_dmatrix = gr.Checkbox(
                                label="XGB Use DMatrix", value=True)
                            xgb_chunk_size = gr.Number(
                                label="XGB Chunk Size (rows, 0=off)", value=0, precision=0)
                            resn_use_lazy_dataset = gr.Checkbox(
                                label="ResNet Lazy Dataset", value=True)
                            resn_predict_batch_size = gr.Number(
                                label="ResNet Predict Batch Size (0=auto)", value=0, precision=0)

                        with gr.Column(scale=1):
                            gr.Markdown("#### Cleanup Controls")
                            xgb_cleanup_per_fold = gr.Checkbox(
                                label="XGB Cleanup Per Fold", value=False)
                            xgb_cleanup_synchronize = gr.Checkbox(
                                label="XGB Cleanup Synchronize", value=False)
                            ft_cleanup_per_fold = gr.Checkbox(
                                label="FT Cleanup Per Fold", value=False)
                            ft_cleanup_synchronize = gr.Checkbox(
                                label="FT Cleanup Synchronize", value=False)
                            resn_cleanup_per_fold = gr.Checkbox(
                                label="ResNet Cleanup Per Fold", value=False)
                            resn_cleanup_synchronize = gr.Checkbox(
                                label="ResNet Cleanup Synchronize", value=False)
                            gnn_cleanup_per_fold = gr.Checkbox(
                                label="GNN Cleanup Per Fold", value=False)
                            gnn_cleanup_synchronize = gr.Checkbox(
                                label="GNN Cleanup Synchronize", value=False)
                            optuna_cleanup_synchronize = gr.Checkbox(
                                label="Optuna Cleanup Synchronize", value=False)

            gr.Markdown("#### Bayesian Optimization Search Spaces (JSON)")
            with gr.Row(equal_height=False):
                with gr.Column(scale=1):
                    xgb_search_space_json = gr.Textbox(
                        label="XGB Search Space",
                        value=xgb_search_space_template,
                        lines=10,
                        max_lines=20,
                    )
                with gr.Column(scale=1):
                    resn_search_space_json = gr.Textbox(
                        label="ResNet Search Space",
                        value=resn_search_space_template,
                        lines=10,
                        max_lines=20,
                    )
            with gr.Row(equal_height=False):
                with gr.Column(scale=1):
                    ft_search_space_json = gr.Textbox(
                        label="FT Supervised Search Space",
                        value=ft_search_space_template,
                        lines=10,
                        max_lines=20,
                    )
                with gr.Column(scale=1):
                    ft_unsupervised_search_space_json = gr.Textbox(
                        label="FT Unsupervised Search Space",
                        value=ft_unsupervised_search_space_template,
                        lines=10,
                        max_lines=20,
                    )

            with gr.Row():
                build_btn = gr.Button(
                    "Build Configuration", variant="primary", size="lg")
                save_config_btn = gr.Button(
                    "Save Configuration", variant="secondary", size="lg")

            build_status = gr.Textbox(label="Status", interactive=False)
            config_json = gr.Textbox(
                label="Generated Config (JSON)", lines=12, max_lines=24)

            with gr.Row(equal_height=True):
                save_filename = gr.Textbox(
                    label="Save Filename", value="my_config.json", scale=3)
                save_status = gr.Textbox(
                    label="Save Status", interactive=False, scale=4)

        with gr.Tab("Run Task"):
            gr.Markdown(
                """
                ### Run Model Task
                Click the button below to execute the task defined in your configuration.
                Task type is automatically detected from `config.runner.mode`:
                - **entry**: Standard model training
                - **explain**: Model explanation (permutation, SHAP, integrated gradients)
                - **incremental**: Incremental training
                - **watchdog**: Watchdog mode

                Task logs will appear in real-time below.
                """
            )

            with gr.Row():
                run_btn = gr.Button("Run Task", variant="primary", size="lg")
                run_status = gr.Textbox(label="Task Status", interactive=False)

            gr.Markdown("### Task Logs")
            log_output = gr.Textbox(
                label="Logs",
                lines=25,
                max_lines=50,
                interactive=False,
                autoscroll=True
            )

            gr.Markdown("---")
            with gr.Row():
                open_folder_btn = gr.Button("Open Results Folder", size="lg")
                folder_status = gr.Textbox(
                    label="Status", interactive=False, scale=2)

        with gr.Tab("FT Two-Step Workflow"):
            gr.Markdown(
                """
                ### FT-Transformer Two-Step Training

                Automates the FT -> XGB/ResN workflow:
                1. **Step 1**: Train FT-Transformer as unsupervised embedding generator
                2. **Step 2**: Merge embeddings with raw data and train XGB/ResN

                **Instructions**:
                1. Load or build a base configuration in the Configuration tab
                2. Prepare Step 1 config (FT embeddings)
                3. Run Step 1 to generate embeddings
                4. Prepare Step 2 configs (XGB/ResN using embeddings)
                5. Run Step 2 with the generated configs
                """
            )

            with gr.Row():
                with gr.Column():
                    gr.Markdown("### Step 1: FT Embedding Generation")
                    ft_use_ddp = gr.Checkbox(
                        label="Use DDP for FT", value=True)
                    ft_nproc = gr.Number(
                        label="Number of Processes (DDP)", value=2, precision=0)

                    prepare_step1_btn = gr.Button(
                        "Prepare Step 1 Config", variant="primary")
                    step1_status = gr.Textbox(
                        label="Status", interactive=False)
                    step1_config_display = gr.Textbox(
                        label="Step 1 Config (FT Embedding)",
                        lines=15,
                        max_lines=25
                    )

                with gr.Column():
                    gr.Markdown("### Step 2: Train XGB/ResN with Embeddings")
                    target_models_input = gr.Textbox(
                        label="Target Models (comma-separated)",
                        value="xgb, resn",
                        placeholder="xgb, resn"
                    )
                    augmented_data_dir_input = gr.Textbox(
                        label="Augmented Data Directory",
                        value="./DataFTUnsupervised",
                        placeholder="./DataFTUnsupervised",
                    )
                    gr.Markdown(
                        "Step-2 parameter overrides (JSON). "
                        "Defaults match `02 Train_FT_Embed_XGBResN.ipynb`."
                    )
                    xgb_overrides_input = gr.Textbox(
                        label="XGB Step 2 Overrides (JSON)",
                        value=xgb_step2_overrides_template,
                        lines=10,
                        max_lines=20,
                    )
                    resn_overrides_input = gr.Textbox(
                        label="ResN Step 2 Overrides (JSON)",
                        value=resn_step2_overrides_template,
                        lines=10,
                        max_lines=20,
                    )

                    prepare_step2_btn = gr.Button(
                        "Prepare Step 2 Configs", variant="primary")
                    step2_status = gr.Textbox(
                        label="Status", interactive=False)

                    with gr.Tab("XGB Config"):
                        xgb_config_display = gr.Textbox(
                            label="XGB Step 2 Config",
                            lines=15,
                            max_lines=25
                        )

                    with gr.Tab("ResN Config"):
                        resn_config_display = gr.Textbox(
                            label="ResN Step 2 Config",
                            lines=15,
                            max_lines=25
                        )

            gr.Markdown("---")
            gr.Markdown(
                """
                ### Quick Actions
                After preparing configs, you can:
                - Copy the Step 1 config and paste it in the **Configuration** tab, then run it in **Run Task** tab
                - After Step 1 completes, click **Prepare Step 2 Configs**
                - Step 2 configs are auto-saved as:
                  - `config_xgb_from_ft_unsupervised.json`
                  - `config_resn_from_ft_unsupervised.json`
                - You can edit XGB/ResN override JSONs before generation to customize Step-2 parameters
                - Run those configs in **Run Task** tab (or keep using the JSON text boxes below)
                """
            )

        with gr.Tab("Workflow Config"):
            gr.Markdown(
                """
                ### Config-Driven Plotting / Prediction / Compare / Pre-Oneway
                Use a JSON config file to run frontend workflows without manual field-by-field input.

                Supported `workflow.mode` values:
                - `pre_oneway`
                - `plot_direct`
                - `plot_embed`
                - `predict_ft_embed`
                - `compare_xgb`
                - `compare_resn`
                - `compare` (requires `model_key`: `xgb` or `resn`)
                - `double_lift`
                """
            )

            with gr.Row():
                with gr.Column(scale=1):
                    workflow_file = gr.File(
                        label="Upload Workflow Config",
                        file_types=[".json"],
                        type="filepath",
                    )
                    workflow_load_btn = gr.Button(
                        "Load Workflow Config", variant="secondary")
                    workflow_load_status = gr.Textbox(
                        label="Load Status", interactive=False)

                with gr.Column(scale=2):
                    workflow_config_json = gr.Textbox(
                        label="Workflow Config (JSON)",
                        value=workflow_template,
                        lines=18,
                        max_lines=30,
                    )

            workflow_run_btn = gr.Button(
                "Run Workflow Config", variant="primary", size="lg")
            workflow_status = gr.Textbox(label="Workflow Status", interactive=False)
            workflow_log = gr.Textbox(
                label="Workflow Logs",
                lines=18,
                max_lines=40,
                interactive=False,
                autoscroll=True,
            )

        with gr.Tab("Plotting"):
            gr.Markdown(
                """
                ### Plotting Workflows
                Run the plotting steps from the example notebooks.
                """
            )

            with gr.Tab("Pre Oneway"):
                with gr.Row():
                    with gr.Column():
                        pre_data_path = gr.Textbox(
                            label="Data Path", value="./Data/od_bc.csv")
                        pre_model_name = gr.Textbox(
                            label="Model Name", value="od_bc")
                        pre_target = gr.Textbox(
                            label="Target Column", value="response")
                        pre_weight = gr.Textbox(
                            label="Weight Column", value="weights")
                        pre_output_dir = gr.Textbox(
                            label="Output Dir (optional)", value="")
                    with gr.Column():
                        pre_feature_list = gr.Textbox(
                            label="Feature List (comma-separated)",
                            lines=4,
                            placeholder="feature_1, feature_2, feature_3",
                        )
                        pre_categorical = gr.Textbox(
                            label="Categorical Features (comma-separated, optional)",
                            lines=3,
                            placeholder="feature_2, feature_3",
                        )
                        pre_n_bins = gr.Number(
                            label="Bins", value=10, precision=0)
                        pre_holdout = gr.Slider(
                            label="Holdout Ratio",
                            minimum=0.0,
                            maximum=0.5,
                            value=0.25,
                            step=0.05,
                        )
                        pre_seed = gr.Number(
                            label="Random Seed", value=13, precision=0)

                pre_run_btn = gr.Button("Run Pre Oneway", variant="primary")
                pre_status = gr.Textbox(label="Status", interactive=False)
                pre_log = gr.Textbox(label="Logs", lines=15,
                                     max_lines=40, interactive=False)

            with gr.Tab("Direct Plot"):
                with gr.Row():
                    with gr.Column(scale=3):
                        direct_cfg_path = gr.Textbox(
                            label="Plot Config", value="config_plot.json")
                        direct_run_btn = gr.Button(
                            "Run Direct Plot", variant="primary")
                        direct_status = gr.Textbox(
                            label="Status", interactive=False)
                    with gr.Column(scale=4):
                        direct_xgb_cfg = gr.Textbox(
                            label="XGB Config", value="config_xgb_direct.json")
                        direct_resn_cfg = gr.Textbox(
                            label="ResN Config", value="config_resn_direct.json")
                direct_log = gr.Textbox(
                    label="Logs", lines=15, max_lines=40, interactive=False)

            with gr.Tab("Embed Plot"):
                with gr.Row():
                    with gr.Column(scale=3):
                        embed_cfg_path = gr.Textbox(
                            label="Plot Config", value="config_plot.json")
                        embed_ft_cfg = gr.Textbox(
                            label="FT Embed Config", value="config_ft_unsupervised_ddp_embed.json")
                        embed_runtime = gr.Checkbox(
                            label="Use Runtime FT Embedding", value=False)
                        embed_run_btn = gr.Button(
                            "Run Embed Plot", variant="primary")
                        embed_status = gr.Textbox(
                            label="Status", interactive=False)
                    with gr.Column(scale=4):
                        embed_xgb_cfg = gr.Textbox(
                            label="XGB Embed Config", value="config_xgb_from_ft_unsupervised.json")
                        embed_resn_cfg = gr.Textbox(
                            label="ResN Embed Config", value="config_resn_from_ft_unsupervised.json")
                embed_log = gr.Textbox(
                    label="Logs", lines=15, max_lines=40, interactive=False)

            with gr.Tab("Double Lift"):
                gr.Markdown(
                    """
                    Draw a double-lift curve from any CSV file with two prediction columns.
                    """
                )
                with gr.Row():
                    with gr.Column():
                        dl_data_path = gr.Textbox(
                            label="Data Path (CSV)", value="./Data/od_bc.csv")
                        dl_pred_col_1 = gr.Textbox(
                            label="Prediction Column 1", value="pred_xgb")
                        dl_pred_col_2 = gr.Textbox(
                            label="Prediction Column 2", value="pred_resn")
                        dl_target_col = gr.Textbox(
                            label="Target Column", value="reponse")
                        dl_weight_col = gr.Textbox(
                            label="Weight Column", value="weights")
                        dl_output_path = gr.Textbox(
                            label="Output Image Path (optional)", value="")
                    with gr.Column():
                        dl_label1 = gr.Textbox(label="Label 1", value="Model 1")
                        dl_label2 = gr.Textbox(label="Label 2", value="Model 2")
                        dl_n_bins = gr.Number(label="Bins", value=10, precision=0)
                        dl_holdout_ratio = gr.Slider(
                            label="Holdout Ratio (0 = all data, >0 = train/test split)",
                            minimum=0.0,
                            maximum=0.5,
                            value=0.0,
                            step=0.05,
                        )
                        dl_split_strategy = gr.Dropdown(
                            label="Split Strategy",
                            choices=["random", "stratified", "time", "group"],
                            value="random",
                        )
                        dl_split_group_col = gr.Textbox(
                            label="Group Column (optional, for group split)",
                            value="",
                        )
                        dl_split_time_col = gr.Textbox(
                            label="Time Column (optional, for time split)",
                            value="",
                        )
                        dl_split_time_ascending = gr.Checkbox(
                            label="Time Ascending",
                            value=True,
                        )
                        dl_rand_seed = gr.Number(
                            label="Random Seed", value=13, precision=0)
                        dl_pred1_weighted = gr.Checkbox(
                            label="Prediction 1 Is Weighted",
                            value=False,
                        )
                        dl_pred2_weighted = gr.Checkbox(
                            label="Prediction 2 Is Weighted",
                            value=False,
                        )
                        dl_actual_weighted = gr.Checkbox(
                            label="Actual Is Weighted",
                            value=False,
                        )

                dl_run_btn = gr.Button("Run Double Lift", variant="primary")
                dl_status = gr.Textbox(label="Status", interactive=False)
                dl_log = gr.Textbox(
                    label="Logs", lines=15, max_lines=40, interactive=False)

            with gr.Tab("FT-Embed Compare"):
                gr.Markdown("Compare Direct vs FT-Embed models and draw double-lift curves.")
                with gr.Row():
                    with gr.Column(scale=4):
                        cmp_model_key = gr.Dropdown(
                            label="Model Key",
                            choices=["xgb", "resn"],
                            value="xgb",
                        )
                        cmp_direct_cfg = gr.Textbox(
                            label="Direct Model Config", value="config_xgb_direct.json")
                        cmp_ft_cfg = gr.Textbox(
                            label="FT Config", value="config_ft_unsupervised_ddp_embed.json")
                        cmp_embed_cfg = gr.Textbox(
                            label="FT-Embed Model Config", value="config_xgb_from_ft_unsupervised.json")
                    with gr.Column(scale=3):
                        cmp_label_direct = gr.Textbox(
                            label="Direct Label", value="XGB_raw")
                        cmp_label_ft = gr.Textbox(
                            label="FT Label", value="XGB_ft_embed")
                        cmp_runtime = gr.Checkbox(
                            label="Use Runtime FT Embedding", value=False)
                        cmp_bins = gr.Number(
                            label="Bins Override", value=10, precision=0)
                        cmp_run_btn = gr.Button("Run Compare", variant="primary")
                        cmp_status = gr.Textbox(label="Status", interactive=False)
                cmp_log = gr.Textbox(
                    label="Logs", lines=15, max_lines=40, interactive=False)

        with gr.Tab("Prediction"):
            gr.Markdown("### FT Embed Prediction")
            with gr.Row():
                with gr.Column(scale=4):
                    pred_ft_cfg = gr.Textbox(
                        label="FT Config", value="config_ft_unsupervised_ddp_embed.json")
                    pred_xgb_cfg = gr.Textbox(
                        label="XGB Config (optional)", value="config_xgb_from_ft_unsupervised.json")
                    pred_resn_cfg = gr.Textbox(
                        label="ResN Config (optional)", value="config_resn_from_ft_unsupervised.json")
                    pred_model_name = gr.Textbox(
                        label="Model Name (optional)", value="")
                    pred_model_keys = gr.Textbox(
                        label="Model Keys", value="xgb, resn")
                with gr.Column(scale=3):
                    pred_input = gr.Textbox(
                        label="Input Data", value="./Data/od_bc_new.csv")
                    pred_output = gr.Textbox(
                        label="Output CSV", value="./Results/predictions_ft_xgb.csv")
                    pred_run_btn = gr.Button("Run Prediction", variant="primary")
                    pred_status = gr.Textbox(label="Status", interactive=False)
            pred_log = gr.Textbox(label="Logs", lines=15,
                                  max_lines=40, interactive=False)

        # Event handlers
        set_working_dir_btn.click(
            fn=_set_working_dir_ui,
            inputs=[working_dir_input],
            outputs=[
                working_dir_status,
                working_dir_input,
                working_dir_browse_root,
                working_dir_picker,
            ],
        )

        refresh_working_dir_btn.click(
            fn=_refresh_working_dir_choices_ui,
            inputs=[working_dir_browse_root],
            outputs=[working_dir_status, working_dir_picker],
        )

        use_selected_working_dir_btn.click(
            fn=_set_working_dir_ui,
            inputs=[working_dir_picker],
            outputs=[
                working_dir_status,
                working_dir_input,
                working_dir_browse_root,
                working_dir_picker,
            ],
        )

        working_dir_picker.change(
            fn=lambda path: str(path or ""),
            inputs=[working_dir_picker],
            outputs=[working_dir_input],
        )

        load_btn.click(
            fn=app.load_json_config,
            inputs=[json_file],
            outputs=[load_status, config_display, config_json]
        )

        build_btn.click(
            fn=app.build_config_from_ui,
            inputs=[
                data_dir, model_list, model_categories, target, weight,
                feature_list, categorical_features, task_type, prop_test,
                holdout_ratio, val_ratio, split_strategy, rand_seed, epochs,
                output_dir, use_gpu, model_keys, max_evals,
                xgb_max_depth_max, xgb_n_estimators_max,
                xgb_gpu_id, xgb_cleanup_per_fold, xgb_cleanup_synchronize,
                xgb_use_dmatrix, xgb_chunk_size,
                xgb_search_space_json, resn_search_space_json,
                ft_search_space_json, ft_unsupervised_search_space_json,
                ft_cleanup_per_fold, ft_cleanup_synchronize,
                resn_cleanup_per_fold, resn_cleanup_synchronize,
                resn_use_lazy_dataset, resn_predict_batch_size,
                gnn_cleanup_per_fold, gnn_cleanup_synchronize,
                optuna_cleanup_synchronize
            ],
            outputs=[build_status, config_json]
        )

        save_config_btn.click(
            fn=app.save_config,
            inputs=[config_json, save_filename],
            outputs=[save_status]
        )

        run_btn.click(
            fn=app.run_training,
            inputs=[config_json],
            outputs=[run_status, log_output]
        )

        open_folder_btn.click(
            fn=app.open_results_folder,
            inputs=[config_json],
            outputs=[folder_status]
        )

        workflow_load_btn.click(
            fn=app.load_workflow_config,
            inputs=[workflow_file],
            outputs=[workflow_load_status, workflow_config_json]
        )

        workflow_run_btn.click(
            fn=app.run_workflow_config_ui,
            inputs=[workflow_config_json],
            outputs=[workflow_status, workflow_log]
        )

        prepare_step1_btn.click(
            fn=app.prepare_ft_step1,
            inputs=[config_json, ft_use_ddp, ft_nproc],
            outputs=[step1_status, step1_config_display]
        )

        prepare_step2_btn.click(
            fn=app.prepare_ft_step2,
            inputs=[gr.State(
                lambda: app.current_step1_config or "temp_ft_step1_config.json"),
                target_models_input,
                augmented_data_dir_input,
                xgb_overrides_input,
                resn_overrides_input,
            ],
            outputs=[step2_status, xgb_config_display, resn_config_display]
        )

        pre_run_btn.click(
            fn=app.run_pre_oneway_ui,
            inputs=[
                pre_data_path, pre_model_name, pre_target, pre_weight,
                pre_feature_list, pre_categorical, pre_n_bins,
                pre_holdout, pre_seed, pre_output_dir
            ],
            outputs=[pre_status, pre_log]
        )

        direct_run_btn.click(
            fn=app.run_plot_direct_ui,
            inputs=[direct_cfg_path, direct_xgb_cfg, direct_resn_cfg],
            outputs=[direct_status, direct_log]
        )

        embed_run_btn.click(
            fn=app.run_plot_embed_ui,
            inputs=[embed_cfg_path, embed_xgb_cfg,
                    embed_resn_cfg, embed_ft_cfg, embed_runtime],
            outputs=[embed_status, embed_log]
        )

        dl_run_btn.click(
            fn=app.run_double_lift_ui,
            inputs=[
                dl_data_path, dl_pred_col_1, dl_pred_col_2, dl_target_col, dl_weight_col,
                dl_n_bins, dl_label1, dl_label2,
                dl_pred1_weighted, dl_pred2_weighted, dl_actual_weighted,
                dl_holdout_ratio, dl_split_strategy, dl_split_group_col, dl_split_time_col,
                dl_split_time_ascending, dl_rand_seed, dl_output_path
            ],
            outputs=[dl_status, dl_log]
        )

        pred_run_btn.click(
            fn=app.run_predict_ui,
            inputs=[
                pred_ft_cfg, pred_xgb_cfg, pred_resn_cfg, pred_input,
                pred_output, pred_model_name, pred_model_keys
            ],
            outputs=[pred_status, pred_log]
        )

        cmp_model_key.change(
            fn=_suggest_compare_defaults,
            inputs=[cmp_model_key],
            outputs=[
                cmp_direct_cfg,
                cmp_embed_cfg,
                cmp_label_direct,
                cmp_label_ft,
            ],
        )

        cmp_run_btn.click(
            fn=app.run_compare_ui,
            inputs=[
                cmp_model_key,
                cmp_direct_cfg,
                cmp_ft_cfg,
                cmp_embed_cfg,
                cmp_label_direct,
                cmp_label_ft,
                cmp_runtime,
                cmp_bins,
            ],
            outputs=[cmp_status, cmp_log]
        )

    return demo


if __name__ == "__main__":
    demo = create_ui()
    launch_kwargs = {
        "server_name": "0.0.0.0",
        "server_port": 7860,
        "share": False,
        "show_error": True,
    }
    if "analytics_enabled" in inspect.signature(demo.launch).parameters:
        launch_kwargs["analytics_enabled"] = False
    demo.launch(**launch_kwargs)



from __future__ import annotations

import importlib.util
import json
from pathlib import Path
import sys
import types

import pandas as pd
import pytest


REPO_ROOT = Path(__file__).resolve().parents[2]
ARTIFACTS_PATH = REPO_ROOT / "modelling" / "bayesopt" / "artifacts.py"
CHECKPOINTS_PATH = REPO_ROOT / "modelling" / "bayesopt" / "checkpoints.py"


def _load_module(module_name: str, module_path: Path):
    spec = importlib.util.spec_from_file_location(module_name, str(module_path))
    if spec is None or spec.loader is None:
        raise RuntimeError(f"Unable to load module from {module_path}")
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


class _FakeIOUtils:
    @staticmethod
    def load_params_file(path: str):
        p = Path(path)
        if p.suffix.lower() == ".json":
            payload = json.loads(p.read_text(encoding="utf-8"))
            if isinstance(payload, dict) and "best_params" in payload:
                payload = payload.get("best_params") or {}
            return dict(payload)
        if p.suffix.lower() in {".csv", ".tsv"}:
            df = pd.read_csv(p, sep="\t" if p.suffix.lower() == ".tsv" else ",")
            if df.empty:
                raise ValueError("empty params file")
            params = df.iloc[0].to_dict()
            return {k: v for k, v in params.items() if not str(k).startswith("Unnamed")}
        raise ValueError("unsupported file type")


def _load_artifacts_module(monkeypatch: pytest.MonkeyPatch):
    fake_io_mod = types.ModuleType("ins_pricing.utils.io")
    fake_io_mod.IOUtils = _FakeIOUtils
    fake_utils_mod = types.ModuleType("ins_pricing.utils")
    fake_utils_mod.io = fake_io_mod
    fake_root_mod = types.ModuleType("ins_pricing")
    fake_root_mod.utils = fake_utils_mod
    monkeypatch.setitem(sys.modules, "ins_pricing", fake_root_mod)
    monkeypatch.setitem(sys.modules, "ins_pricing.utils", fake_utils_mod)
    monkeypatch.setitem(sys.modules, "ins_pricing.utils.io", fake_io_mod)
    return _load_module("test_artifacts_module", ARTIFACTS_PATH)


def test_best_params_filename_and_model_key_mapping(monkeypatch: pytest.MonkeyPatch):
    artifacts = _load_artifacts_module(monkeypatch)
    assert artifacts.best_params_filename("demo", "ResNet") == "demo_bestparams_resnet.csv"
    assert artifacts.trainer_label_from_model_key("FT") == "fttransformer"
    assert artifacts.trainer_label_from_model_key("custom") == "custom"


def test_load_best_params_prefers_version_snapshot(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    artifacts = _load_artifacts_module(monkeypatch)
    result_dir = tmp_path / "Results"
    versions_dir = result_dir / "versions"
    versions_dir.mkdir(parents=True)

    (versions_dir / "20260101_120000_resn_best.json").write_text(
        json.dumps({"best_params": {"hidden_dim": 32}}),
        encoding="utf-8",
    )
    (result_dir / "demo_bestparams_resnet.csv").write_text(
        "hidden_dim,dropout\n64,0.1\n",
        encoding="utf-8",
    )

    loaded = artifacts.load_best_params(tmp_path, "demo", "resn")
    assert loaded == {"hidden_dim": 32}


def test_load_best_params_falls_back_to_csv(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    artifacts = _load_artifacts_module(monkeypatch)
    result_dir = tmp_path / "Results"
    result_dir.mkdir(parents=True)
    (result_dir / "demo_bestparams_resnet.csv").write_text(
        "hidden_dim,dropout\n64,0.2\n",
        encoding="utf-8",
    )

    loaded = artifacts.load_best_params(tmp_path, "demo", "resn")
    assert loaded == {"hidden_dim": 64.0, "dropout": 0.2}


def test_rebuild_ft_model_from_payload_merges_missing_config(monkeypatch: pytest.MonkeyPatch):
    checkpoints = _load_module("test_checkpoints_module_ft", CHECKPOINTS_PATH)
    captured = {}

    def _fake_rebuild_ft_model_from_checkpoint(*, state_dict, model_config):
        captured["state_dict"] = state_dict
        captured["model_config"] = dict(model_config)
        return {"rebuilt": True}

    monkeypatch.setattr(
        checkpoints,
        "rebuild_ft_model_from_checkpoint",
        _fake_rebuild_ft_model_from_checkpoint,
    )
    payload = {
        "state_dict": {"weights": [1, 2, 3]},
        "model_config": {"loss_name": "poisson", "distribution": None},
        "best_params": {"n_heads": 4},
    }

    model, best_params, kind = checkpoints.rebuild_ft_model_from_payload(
        payload=payload,
        model_config_overrides={"loss_name": "gamma", "distribution": "gamma"},
        fill_missing_model_config=True,
    )

    assert model == {"rebuilt": True}
    assert best_params == {"n_heads": 4}
    assert kind == "state_dict"
    assert captured["state_dict"] == {"weights": [1, 2, 3]}
    assert captured["model_config"]["loss_name"] == "poisson"
    assert captured["model_config"]["distribution"] == "gamma"


def test_rebuild_resn_model_from_payload_with_fallback_params():
    checkpoints = _load_module("test_checkpoints_module_resn", CHECKPOINTS_PATH)

    class _ResNetCore:
        def __init__(self):
            self.loaded = None

        def load_state_dict(self, state_dict):
            self.loaded = state_dict

    class _Model:
        def __init__(self, params):
            self.params = dict(params)
            self.resnet = _ResNetCore()

    def _builder(params):
        return _Model(params)

    model, params = checkpoints.rebuild_resn_model_from_payload(
        payload={"state_dict": {"layer": [1, 2]}},
        model_builder=_builder,
        params_fallback={"hidden_dim": 64},
    )
    assert params == {"hidden_dim": 64}
    assert model.params == {"hidden_dim": 64}
    assert model.resnet.loaded == {"layer": [1, 2]}


def test_rebuild_gnn_model_from_payload_strict_fallback_warning():
    checkpoints = _load_module("test_checkpoints_module_gnn", CHECKPOINTS_PATH)

    class _GNNCore:
        def __init__(self):
            self.calls = []

        def load_state_dict(self, state_dict, strict=True):
            self.calls.append((state_dict, strict))
            if strict:
                raise RuntimeError("Missing key(s) in state_dict")

    class _Model:
        def __init__(self, params):
            self.params = dict(params)
            self._core = _GNNCore()

        def set_params(self, params):
            self.params = dict(params)

        def _unwrap_gnn(self):
            return self._core

    def _builder(params):
        return _Model(params)

    model, params, warning = checkpoints.rebuild_gnn_model_from_payload(
        payload={"best_params": {"k_neighbors": 10}, "state_dict": {"w": [1]}},
        model_builder=_builder,
        strict=True,
        allow_non_strict_fallback=True,
    )

    assert params == {"k_neighbors": 10}
    assert model.params == {"k_neighbors": 10}
    assert warning is not None
    assert model._core.calls == [({"w": [1]}, True), ({"w": [1]}, False)]


def test_rebuild_gnn_model_from_payload_requires_dict_payload():
    checkpoints = _load_module("test_checkpoints_module_gnn_err", CHECKPOINTS_PATH)
    with pytest.raises(ValueError, match="Invalid GNN checkpoint payload"):
        checkpoints.rebuild_gnn_model_from_payload(
            payload="bad",
            model_builder=lambda params: object(),
        )




from __future__ import annotations

from typing import Iterable, List, Mapping, Optional, Sequence, Tuple

import numpy as np
import pandas as pd

from ins_pricing.modelling.plotting.common import PlotStyle, plt
from ins_pricing.modelling.plotting.curves import (
    plot_double_lift_curve,
    plot_lift_curve,
)


def _safe_tag(value: str) -> str:
    return "".join(ch if ch.isalnum() or ch in "-_." else "_" for ch in str(value))


def _to_array(values: Sequence[float], name: str) -> np.ndarray:
    if values is None:
        raise ValueError(f"{name} is required.")
    return np.asarray(values, dtype=float).reshape(-1)


def _resolve_pred_pairs_from_df(
    df: pd.DataFrame,
    pred_cols: Optional[Sequence[str] | Mapping[str, object]],
    pred_labels: Optional[Sequence[str]],
) -> List[Tuple[str, np.ndarray]]:
    if pred_cols is None:
        raise ValueError("pred_cols is required when data is a DataFrame.")

    pairs: List[Tuple[str, np.ndarray]] = []
    if isinstance(pred_cols, Mapping):
        for label, col in pred_cols.items():
            if isinstance(col, str) and col in df.columns:
                arr = df[col].to_numpy()
            else:
                arr = np.asarray(col, dtype=float).reshape(-1)
            pairs.append((str(label), arr))
        return pairs

    if isinstance(pred_cols, str):
        label = pred_cols if not pred_labels else str(pred_labels[0])
        return [(label, df[pred_cols].to_numpy())]

    labels = list(pred_labels) if pred_labels else [str(c) for c in pred_cols]
    for col, label in zip(pred_cols, labels):
        if isinstance(col, str) and col in df.columns:
            arr = df[col].to_numpy()
        else:
            arr = np.asarray(col, dtype=float).reshape(-1)
        pairs.append((str(label), arr))
    return pairs


def _resolve_pred_pairs_from_arrays(
    preds: object,
    pred_labels: Optional[Sequence[str]],
) -> List[Tuple[str, np.ndarray]]:
    if isinstance(preds, Mapping):
        return [(str(k), np.asarray(v, dtype=float).reshape(-1)) for k, v in preds.items()]

    if isinstance(preds, (list, tuple)):
        labels = list(pred_labels) if pred_labels else [f"model_{idx + 1}" for idx in range(len(preds))]
        return [
            (str(label), np.asarray(arr, dtype=float).reshape(-1))
            for label, arr in zip(labels, preds)
        ]

    return [("model_1", np.asarray(preds, dtype=float).reshape(-1))]


def plot_lift_list(
    data_or_preds: pd.DataFrame | Mapping[str, object] | Sequence[object],
    pred_cols: Optional[Sequence[str] | Mapping[str, object]] = None,
    actual: Optional[Sequence[float]] = None,
    weight: Optional[Sequence[float]] = None,
    *,
    actual_col: str = "w_act",
    weight_col: str = "weight",
    pred_labels: Optional[Sequence[str]] = None,
    n_bins: int = 10,
    pred_weighted: bool = False,
    actual_weighted: bool = True,
    show: bool = False,
    save_dir: Optional[str] = None,
    filename_prefix: str = "lift",
    style: Optional[PlotStyle] = None,
) -> List[plt.Figure]:
    """Compatibility helper: plot multiple lift curves in one call."""
    if isinstance(data_or_preds, pd.DataFrame):
        df = data_or_preds
        pairs = _resolve_pred_pairs_from_df(df, pred_cols, pred_labels)
        actual_arr = _to_array(df[actual_col], actual_col) if actual is None else _to_array(actual, "actual")
        weight_arr = None
        if weight is not None:
            weight_arr = _to_array(weight, "weight")
        elif weight_col in df.columns:
            weight_arr = _to_array(df[weight_col], weight_col)
    else:
        pairs = _resolve_pred_pairs_from_arrays(data_or_preds, pred_labels)
        actual_arr = _to_array(actual, "actual")
        weight_arr = _to_array(weight, "weight") if weight is not None else None

    figs: List[plt.Figure] = []
    for label, pred in pairs:
        save_path = None
        if save_dir:
            safe_dir = save_dir.rstrip("/\\")
            save_path = f"{safe_dir}/{filename_prefix}_{_safe_tag(label)}.png"
        fig = plot_lift_curve(
            pred,
            actual_arr,
            weight_arr,
            n_bins=n_bins,
            title=f"{label} Lift Chart",
            pred_label="Predicted",
            act_label="Actual",
            weight_label="Earned Exposure",
            pred_weighted=pred_weighted,
            actual_weighted=actual_weighted,
            show=show,
            save_path=save_path,
            style=style,
        )
        figs.append(fig)
    return figs


def plot_dlift_list(
    data_or_preds: pd.DataFrame | Mapping[str, object] | Sequence[object],
    pred_cols: Optional[Sequence[str] | Mapping[str, object]] = None,
    actual: Optional[Sequence[float]] = None,
    weight: Optional[Sequence[float]] = None,
    *,
    actual_col: str = "w_act",
    weight_col: str = "weight",
    pred_labels: Optional[Sequence[str]] = None,
    pairs: Optional[Iterable[Tuple[object, object]]] = None,
    n_bins: int = 10,
    pred_weighted: bool = False,
    actual_weighted: bool = True,
    show: bool = False,
    save_dir: Optional[str] = None,
    filename_prefix: str = "double_lift",
    style: Optional[PlotStyle] = None,
) -> List[plt.Figure]:
    """Compatibility helper: plot double-lift curves for multiple model pairs."""
    if isinstance(data_or_preds, pd.DataFrame):
        df = data_or_preds
        pairs_list = _resolve_pred_pairs_from_df(df, pred_cols, pred_labels)
        actual_arr = _to_array(df[actual_col], actual_col) if actual is None else _to_array(actual, "actual")
        weight_arr = None
        if weight is not None:
            weight_arr = _to_array(weight, "weight")
        elif weight_col in df.columns:
            weight_arr = _to_array(df[weight_col], weight_col)
    else:
        pairs_list = _resolve_pred_pairs_from_arrays(data_or_preds, pred_labels)
        actual_arr = _to_array(actual, "actual")
        weight_arr = _to_array(weight, "weight") if weight is not None else None

    pred_map = {label: arr for label, arr in pairs_list}
    labels = [label for label, _ in pairs_list]

    pair_labels: List[Tuple[str, str]] = []
    if pairs is None:
        for idx, first in enumerate(labels):
            for second in labels[idx + 1 :]:
                pair_labels.append((first, second))
    else:
        for left, right in pairs:
            if isinstance(left, int):
                left_label = labels[left]
            else:
                left_label = str(left)
            if isinstance(right, int):
                right_label = labels[right]
            else:
                right_label = str(right)
            pair_labels.append((left_label, right_label))

    figs: List[plt.Figure] = []
    for label1, label2 in pair_labels:
        pred1 = pred_map.get(label1)
        pred2 = pred_map.get(label2)
        if pred1 is None or pred2 is None:
            continue
        save_path = None
        if save_dir:
            safe_dir = save_dir.rstrip("/\\")
            tag = f"{_safe_tag(label1)}_vs_{_safe_tag(label2)}"
            save_path = f"{safe_dir}/{filename_prefix}_{tag}.png"
        fig = plot_double_lift_curve(
            pred1,
            pred2,
            actual_arr,
            weight_arr,
            n_bins=n_bins,
            title=f"Double Lift: {label1} vs {label2}",
            label1=label1,
            label2=label2,
            pred1_weighted=pred_weighted,
            pred2_weighted=pred_weighted,
            actual_weighted=actual_weighted,
            show=show,
            save_path=save_path,
            style=style,
        )
        figs.append(fig)
    return figs


class PlotUtils:
    """Compatibility wrapper for legacy plotting helpers."""

    plot_lift_curve = staticmethod(plot_lift_curve)
    plot_double_lift_curve = staticmethod(plot_double_lift_curve)
    plot_lift_list = staticmethod(plot_lift_list)
    plot_dlift_list = staticmethod(plot_dlift_list)


__all__ = [
    "PlotUtils",
    "plot_lift_list",
    "plot_dlift_list",
]
